Type: Package
Package: GeoThinneR
Title: Efficient Spatial Thinning of Species Occurrences
Version: 2.1.2
Authors@R: 
    person("Jorge", "Mestre-Tomás", , "jorge.mestre.tomas@csic.es", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-8983-3417"))
Description: Provides efficient geospatial thinning algorithms to reduce
    the density of coordinate data while maintaining spatial
    relationships. Implements K-D Tree and brute-force distance-based
    thinning, as well as grid-based and precision-based thinning methods.
    For more information on the methods, see Elseberg et al. (2012)
    <https://hdl.handle.net/10446/86202>.
License: MIT + file LICENSE
URL: https://github.com/jmestret/GeoThinneR,
        https://jmestret.github.io/GeoThinneR/
BugReports: https://github.com/jmestret/GeoThinneR/issues
Depends: R (>= 4.0.0)
Imports: data.table, doParallel, fields, foreach, graphics, grDevices,
        matrixStats, nabor, sf, stats, terra
Suggests: ggplot2, knitr, rmarkdown, s2 (>= 1.1.0), testthat (>=
        3.0.0), tibble
VignetteBuilder: knitr
BuildVignettes: true
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
LazyData: true
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-08-25 17:20:26 UTC; jorge
Author: Jorge Mestre-Tomás [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-8983-3417>)
Maintainer: Jorge Mestre-Tomás <jorge.mestre.tomas@csic.es>
Repository: CRAN
Date/Publication: 2026-08-25 22:00:14 UTC

# Generated by roxygen2: do not edit by hand

S3method(as_sf,GeoThinned)
S3method(get_trial,GeoThinned)
S3method(largest,GeoThinned)
S3method(largest_index,GeoThinned)
S3method(plot,GeoThinned)
S3method(print,GeoThinned)
S3method(print,summary.GeoThinned)
S3method(summary,GeoThinned)
export(as_GeoThinned)
export(as_sf)
export(calculate_spatial_coverage)
export(compute_nearest_neighbor_distances)
export(compute_neighbors_brute)
export(compute_neighbors_kdtree)
export(compute_neighbors_local_kdtree)
export(distance_thinning)
export(estimate_k_max)
export(get_trial)
export(grid_thinning)
export(is_lonlat)
export(largest)
export(largest_index)
export(lon_lat_to_cartesian)
export(max_thinning_algorithm)
export(new_GeoThinned)
export(precision_thinning)
export(select_target_points)
export(thin_points)
importFrom(data.table,":=")
importFrom(data.table,.N)
importFrom(foreach,"%dopar%")
importFrom(graphics,legend)
importFrom(graphics,points)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# GeoThinneR - An R Package for Efficient Spatial Thinning of Species Occurrences and Point Data <a href="https://jmestret.github.io/GeoThinneR/"><img src="man/figures/logo.png" align="right" height="138" /></a>

<!-- badges: start -->

[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/GeoThinneR?color=blue)](https://cran.r-project.org/package=GeoThinneR)
[![R-CMD-check](https://github.com/jmestret/GeoThinneR/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/jmestret/GeoThinneR/actions/workflows/R-CMD-check.yaml)
[![codecov](https://codecov.io/gh/jmestret/GeoThinneR/graph/badge.svg?token=LID8Q55SMD)](https://app.codecov.io/gh/jmestret/GeoThinneR)
<!-- badges: end -->

## Overview

**GeoThinneR** is an R package designed for efficient spatial thinning
of species occurrence records and other geospatial point data. It
integrates three primary thinning methods (**distance-based, grid-based,
and precision-based thinning**) into a single package, eliminating the
need to switch between multiple tools. GeoThinneR implements algorithms
based on kd-tree structures for nearest-neighbor searches, significantly
improving performance and **scalability for large datasets**.
Additionally, the package provides custom **functionalities** useful for
species distribution modeling (SDM), such as thinning by group (e.g.,
multiple species), retaining an exact number of points, and prioritizing
records based on user-defined variables. These features make GeoThinneR
a valuable tool for handling large-scale occurrence datasets.

<img src="man/figures/README-GeoThinneR_summary.png" width="90%" />

GeoThinneR has been developed as an alternative tool for spatial
thinning to mitigate the effects of sampling bias in SDM. Various
approaches exist to address sampling bias, each suited to different
scenarios. Below are some references discussing methods for bias
correction and spatial thinning:

- Boria, R. A., Olson, L. E., Goodman, S. M., & Anderson, R. P. (2014).
  Spatial filtering to reduce sampling bias can improve the performance
  of ecological niche models. *Ecological modelling, 275*, 73-77.
  <https://doi.org/10.1016/j.ecolmodel.2013.12.012>
- Veloz, S. D. (2009). Spatially autocorrelated sampling falsely
  inflates measures of accuracy for presence‐only niche models. *Journal
  of biogeography, 36*(12),
  2290-2299.<https://doi.org/10.1111/j.1365-2699.2009.02174.x>
- Moudrý, V., Bazzichetto, M., Remelgado, R., Devillers, R., Lenoir, J.,
  Mateo, R.G., Lembrechts, J.J., Sillero, N., Lecours, V., Cord, A.F.,
  Barták, V., Balej, P., Rocchini, D., Torresani, M., Arenas-Castro, S.,
  Man, M., Prajzlerová, D., Gdulová, K., Prošek, J., Marchetto, E.,
  Zarzo-Arias, A., Gábor, L., Leroy, F., Martini, M., Malavasi, M.,
  Cazzolla Gatti, R., Wild, J. and Šímová, P. (2024), Optimising
  occurrence data in species distribution models: sample size,
  positional uncertainty, and sampling bias matter. *Ecography, 2024*:
  e07294. <https://doi.org/10.1111/ecog.07294>

## Getting started

You can install **GeoThinneR** from CRAN with:

``` r
install.packages("GeoThinneR")
```

To install the **development version** from GitHub, use:

``` r
# install.packages("devtools")
devtools::install_github("jmestret/GeoThinneR")
```

Using GeoThinneR is simple. The main function, `thin_points()`, applies
spatial thinning using a user-specified method and thinning constraint.

``` r
library(GeoThinneR)

# Distance-based thinning (minimum separation of 10 km)
thin_points(data, method = "distance", thin_dist = 10)

# Grid-based thinning (grid resolution of 0.1 degrees)
thin_points(data, method = "grid", resolution = 0.1)

# Precision-based thinning (rounding coordinates to 1 decimal place)
thin_points(data, method = "precision", precision = 1)
```

## Documentation

For detailed documentation, guides, and usage examples, please visit the
[official package
documentation](https://jmestret.github.io/GeoThinneR/).

## Contributing

We welcome contributions! If you have suggestions for improvements or
new features, please open an issue or submit a pull request on our
[GitHub repository](https://github.com/jmestret/GeoThinneR).

## How to cite GeoThinneR

If you use GeoThinneR in your work, please cite the published article:

> Mestre-Tomás, J. (2026). GeoThinneR: An R Package for Efficient
> Spatial Thinning of Species Occurrences and Point Data. *The R
> Journal, 18*(1), 299-314. <https://doi.org/10.32614/RJ-2026-006>

## ----include = FALSE----------------------------------------------------------
run_code <- requireNamespace("sf", quietly = TRUE) &
  requireNamespace("terra", quietly = TRUE) &
  requireNamespace("ggplot2", quietly = TRUE)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = run_code
)

## ----setup, message=FALSE, warning=FALSE--------------------------------------
library(GeoThinneR)
library(terra)
library(sf)
library(ggplot2)

## ----load data, message=FALSE, results='hide'---------------------------------
# Set seed for reproducibility
set.seed(123)

# Simulated dataset
n <- 2000
sim_data <- data.frame(
  lon = runif(n, min = -10, max = 10),
  lat = runif(n, min = -5, max = 5),
  sp = sample(c("sp1", "sp2"), n, replace = TRUE)
)

# Load real-world occurrence dataset
data("caretta")

# Load Mediterranean Sea polygon
medit <- system.file("extdata", "mediterranean_sea.gpkg", package = "GeoThinneR")
medit <- sf::st_read(medit)

## ----plot loaded data, fig.show="hold", out.width="50%", echo = FALSE---------
ggplot() +
  geom_point(data = sim_data, aes(x = lon, y = lat, shape = sp), color = "#5183B3") +
  scale_shape_manual(values = c(sp1 = 16, sp2 = 17)) +
  xlab("Longitude") + ylab("Latitude") +
  ggtitle("Simulated Species Occurrences") +
  theme_minimal()

ggplot() +
  geom_sf(data = medit, color = "#353839", fill = "antiquewhite", alpha = 0.7)  +
  geom_point(data = caretta, aes(x = decimalLongitude, y = decimalLatitude), color = "#5183B3", alpha = 1) +
  xlab("Longitude") + ylab("Latitude") +
  ggtitle("C. caretta Mediterranean Sea Occurrences") +
  theme(
      panel.grid.major = element_line(color = gray(.5),linetype = "dashed",linewidth = 0.5),
      panel.background = element_rect(fill = "white"),
      axis.title = element_blank(),
      legend.position = "bottom"
    )

## -----------------------------------------------------------------------------
# Apply spatial thinning to the simulated data
quick_thin <- thin_points(
  data = sim_data,     # Dataframe with coordinates
  lon_col = "lon",     # Longitude column name
  lat_col = "lat",     # Latitude column name
  method = "distance", # Method for thinning
  thin_dist = 20,      # Thinning distance in km,
  trials = 5,          # Number of trials
  all_trials = TRUE,   # Return all trials
  seed = 123           # Seed for reproducibility
)

quick_thin

## -----------------------------------------------------------------------------
summary(quick_thin)

## -----------------------------------------------------------------------------
plot(quick_thin)

## -----------------------------------------------------------------------------
# Number of kept points in each trial
sapply(quick_thin$retained, sum)

## -----------------------------------------------------------------------------
head(largest(quick_thin))

## -----------------------------------------------------------------------------
head(get_trial(quick_thin, trial = 2)) 

## -----------------------------------------------------------------------------
trial <- 2
head(quick_thin$original_data[!quick_thin$retained[[trial]], ])

## -----------------------------------------------------------------------------
sf_thinned <- as_sf(quick_thin)
class(sf_thinned)

## -----------------------------------------------------------------------------
# Apply spatial thinning to the real data
caretta_thin <- thin_points(
  data = caretta, # We will not specify lon_col, lat_col as they are in position 1 and 2
  lat_col = "decimalLatitude",
  lon_col = "decimalLongitude",
  method = "distance",
  thin_dist = 30, # Thinning distance in km,
  trials = 5,
  all_trials = FALSE,
  seed = 123
)

# Thinned dataframe stored in the first element of the retained list
identical(largest(caretta_thin), get_trial(caretta_thin, 1))

## ----echo = FALSE-------------------------------------------------------------
ggplot() +
  geom_sf(data = medit, color = "#353839", fill = "antiquewhite", alpha = 0.7)  +
  geom_point(data = caretta, aes(x = decimalLongitude, y = decimalLatitude),color = "#EB714B", alpha = 0.5) +
  geom_point(data = largest(caretta_thin), aes(x = decimalLongitude, y = decimalLatitude),color = "#5183B3", alpha = 1) +
  xlab("Longitude") + ylab("Latitude") +
  ggtitle("C. caretta Mediterranean Sea Occurrences") +
  theme(
      panel.grid.major = element_line(color = gray(.5),linetype = "dashed",linewidth = 0.5),
      panel.background = element_rect(fill = "white"),
      axis.title = element_blank(),
      legend.position = "bottom"
    )

## -----------------------------------------------------------------------------
dist_thin <- thin_points(sim_data, method = "distance", thin_dist = 25, trials = 3)
plot(dist_thin)

## -----------------------------------------------------------------------------
brute_thin <- thin_points(
  data = sim_data,
  method = "distance",
  search_type = "brute",
  thin_dist = 20,
  trials = 10,
  all_trials = FALSE,
  seed = 123
)
nrow(largest(brute_thin))

## -----------------------------------------------------------------------------
kd_tree_thin <- thin_points(
  data = sim_data,
  method = "distance",
  search_type = "kd_tree",
  thin_dist = 20,
  trials = 10,
  all_trials = FALSE,
  seed = 123
)
nrow(largest(kd_tree_thin))

## -----------------------------------------------------------------------------
local_kd_tree_thin <- thin_points(
  data = sim_data,
  method = "distance",
  search_type = "local_kd_tree",
  thin_dist = 20,
  trials = 10,
  all_trials = FALSE,
  seed = 123
)
nrow(largest(local_kd_tree_thin))

## -----------------------------------------------------------------------------
k_estimation_thin <- thin_points(
  data = sim_data,
  method = "distance",
  search_type = "k_estimation",
  thin_dist = 20,
  trials = 10,
  all_trials = FALSE,
  seed = 123
)
nrow(largest(k_estimation_thin))

## -----------------------------------------------------------------------------
system.time(
grid_thin <- thin_points(
  data = sim_data,
  method = "grid",
  resolution = 1,
  origin = c(0, 0),
  crs = "epsg:4326",
  n = 1, # one point per grid cell
  trials = 10,
  all_trials = FALSE,
  seed = 123
))
nrow(largest(grid_thin))

## ----message=FALSE, warning=FALSE---------------------------------------------
rast_obj <- terra::rast(xmin = -10, xmax = 10, ymin = -5, ymax = 5, res = 1)
system.time(
grid_raster_thin <- thin_points(
  data = sim_data,
  method = "grid", 
  raster_obj = rast_obj,
  trials = 10,
  n = 1,
  all_trials = FALSE,
  seed = 123
))
nrow(largest(grid_raster_thin))

## ----echo = FALSE-------------------------------------------------------------
rast_obj <- terra::as.polygons(rast_obj)
rast_obj <- sf::st_as_sf(rast_obj)
ggplot() +
  geom_sf(data = rast_obj, color = "#353839") +
  geom_point(data = sim_data, aes(x = lon, y = lat),color = "#EB714B", alpha = 0.5) +
  geom_point(data = largest(grid_raster_thin), aes(x = lon, y = lat),color = "#5183B3", alpha = 1) +
  labs(title = "Grid-based thinning results",
       x = "Longitude",
       y = "Latitude",
       caption = "Blue points are kept, red points are deleted") +
  theme_minimal()

## -----------------------------------------------------------------------------
system.time(
precision_thin <- thin_points(
  data = sim_data,
  method = "precision", 
  precision = 0,
  trials = 50,
  all_trials = FALSE,
  seed = 123
))
nrow(largest(precision_thin))

## ----echo = FALSE-------------------------------------------------------------
ggplot() +
  geom_point(data = sim_data, aes(x = lon, y = lat),  color = "#EB714B", size = 3, alpha = 0.5) +
  geom_point(data = largest(precision_thin), aes(x = lon, y = lat), color = "#5183B3", size = 3) +
  labs(title = "Precision-based thinning results",
       x = "Longitude",
       y = "Latitude",
       caption = "Blue points are kept, red points are deleted") +
  theme_minimal()

## -----------------------------------------------------------------------------
all_thin <- thin_points(
  data = sim_data,
  thin_dist = 100,
  seed = 123
)
grouped_thin <- thin_points(
  data = sim_data,
  group_col = "sp",
  thin_dist = 100,
  seed = 123
)

nrow(largest(all_thin))
nrow(largest(grouped_thin))

## ----echo =FALSE, fig.show="hold",  out.width="50%"---------------------------
removed <- sim_data[!all_thin$retained[[1]], ]
removed_group <- sim_data[!grouped_thin$retained[[1]], ]

ggplot() +
  geom_point(data = removed, aes(x = lon, y = lat, shape = sp), color = "#EB714B",  alpha = 0.2) +
  geom_point(data = largest(all_thin), aes(x = lon, y = lat, shape = sp), color = "#5183B3", alpha = 1) +
  scale_shape_manual(values = c(sp1 = 15, sp2 = 17)) +
  ggtitle("Thinning without grouping") +
  theme_minimal()

ggplot() +
  geom_point(data = removed_group, aes(x = lon, y = lat, shape = sp), color = "#EB714B",  alpha = 0.2) +
  geom_point(data = largest(grouped_thin), aes(x = lon, y = lat, shape = sp), color = "#5183B3", alpha = 1) +
  scale_shape_manual(values = c(sp1 = 15, sp2 = 17)) +
  ggtitle("Thinning by group") +
  theme_minimal()

## -----------------------------------------------------------------------------
targeted_thin <- thin_points(
  data = caretta,
  lon_col = "decimalLongitude",
  lat_col = "decimalLatitude",
  target_points = 150,
  thin_dist = 30,
  all_trials = FALSE,
  seed = 123,
  verbose = TRUE
)
nrow(largest(targeted_thin))

## ----echo = FALSE-------------------------------------------------------------
ggplot() +
  geom_sf(data = medit, color = "#353839", fill = "antiquewhite", alpha = 0.7)  +
  geom_point(data = largest(targeted_thin), aes(x = decimalLongitude, y = decimalLatitude),color = "#5183B3", alpha = 1) +
  xlab("Longitude") + ylab("Latitude") +
  ggtitle("Exact number of occurrences") +
  theme(
      panel.grid.major = element_line(color = gray(.5),linetype = "dashed",linewidth = 0.5),
      panel.background = element_rect(fill = "white"),
      axis.title = element_blank(),
      legend.position = "bottom"
    )

## -----------------------------------------------------------------------------
# Invert uncertainty to give more priority to lower uncertainty records
priority <- -caretta$coordinateUncertaintyInMeters

# For example, handle NAs by assigning lowest possible priority
priority[is.na(priority)] <- min(priority, na.rm = TRUE) - 1

## -----------------------------------------------------------------------------
# Standard grid thinning (no priority)
grid_thin <- thin_points(
  data = caretta,
  lon_col = "decimalLongitude",
  lat_col = "decimalLatitude",
  method = "grid",
  resolution = 2,
  seed = 123
)

# Grid thinning with priority
priority_thin <- thin_points(
  data = caretta,
  lon_col = "decimalLongitude",
  lat_col = "decimalLatitude",
  method = "grid",
  resolution = 2,
  priority = priority,
  seed = 123
)

mean(largest(grid_thin)$coordinateUncertaintyInMeters, na.rm = TRUE)
mean(largest(priority_thin)$coordinateUncertaintyInMeters, na.rm = TRUE)

## -----------------------------------------------------------------------------
sessionInfo()

#' Precision Thinning of Spatial Points
#'
#' This function performs thinning of spatial points by rounding their coordinates to a specified precision and removing duplicates.
#' It can perform multiple trials of this process and return the results for all or just the best trial.
#'
#' @param coordinates A numeric matrix or data frame with two columns representing the longitude and latitude of points.
#' @param precision A positive integer specifying the number of decimal places to which coordinates should be rounded. Default is 4.
#' @param trials A positive integer specifying the number of thinning trials to perform. Default is 10.
#' @param all_trials A logical value indicating whether to return results for all trials (`TRUE`) or just the first/best trial (`FALSE`). Default is `FALSE`.
#' @param priority A numeric vector of the same length as the number of points, specifying a priority weight for each point. Higher values indicate higher importance and are favored when selecting which points to retain. Priority is used to guide selection when multiple candidate points are otherwise equally valid (e.g., points in the same grid cell, with the same rounded coordinates, or with the same number of neighbors).
#' @return If `all_trials` is `FALSE`, returns a logical vector indicating which points were kept in the first trial.
#' If `all_trials` is `TRUE`, returns a list of logical vectors, one for each trial.
#'
#' @details
#' The function performs multiple trials to account for randomness in the order of point selection. By default, it returns the first trial,
#' but setting `all_trials = TRUE` will return the results of all trials.
#'
#' @examples
#' # Example usage
#' coords <- matrix(c(-123.3656, 48.4284, -123.3657, 48.4285, -123.3658, 48.4286), ncol = 2)
#' result <- precision_thinning(coords, precision = 3, trials = 5, all_trials = TRUE)
#' print(result)
#'
#' # Example with a single trial and lower precision
#' result_single <- precision_thinning(coords, precision = 2, trials = 1, all_trials = FALSE)
#' print(result_single)
#'
#' @importFrom data.table := .N
#' @export
precision_thinning <- function(coordinates, precision = 4, trials = 10, all_trials = FALSE, priority = NULL) {
  # Validate inputs
  if (!is.numeric(precision) || length(precision) != 1 || precision < 0) {
    stop("`precision` must be a non-negative integer.")
  }

  if (!is.null(priority)){
    if (!is.numeric(priority) | length(priority) != nrow(coordinates)){
      stop("'priority' must be a numeric vector with same length as number of points.")
    }

    if (any(is.na(priority))){
      warning("NA values found in 'priority'. Replacing with lowest priority (-Inf).")
      priority[is.na(priority)] <- -Inf
    }
  }

  # Initialize results list for trials
  keep_points <- vector("list", ifelse(all_trials, trials, 1))

  # Main thinning loop for the specified number of trials
  for (i in seq_len(length(keep_points))) {
    # Create a data.table with rounded coordinates and a random order or by priority
    if (is.null(priority)){
      sort_order <- stats::runif(nrow(coordinates))
    } else {
      sort_order <- rank(-priority, ties.method = "random")
    }
    keep_points_trial <- data.table::data.table(
      id = seq_len(nrow(coordinates)),
      lon = round(coordinates[, 1], precision),
      lat = round(coordinates[, 2], precision),
      rand_order = sort_order
    )

    # Randomize the order to avoid bias
    keep_points_trial <- keep_points_trial[order(keep_points_trial[["rand_order"]], decreasing = TRUE), ]

    # Mark unique points based on rounded coordinates
    keep_points_trial[, "keep"] <- !duplicated(keep_points_trial[, c("lon", "lat")])

    # Sort back to the original order and store the "keep" vector
    keep_points[[i]] <- keep_points_trial[order(keep_points_trial[["id"]]), ][["keep"]]
}

  return(keep_points)
}

coordinates <- matrix(c(-123.3656, -123.3657, -123.3658, 48.4284, 48.4285, 48.4286), ncol = 2)
priority <- 1:3

test_that("precision_thinning basic functionality with single trial", {
  result <- precision_thinning(coordinates)

  expect_true(is.logical(result[[1]]))
  expect_equal(length(result[[1]]), nrow(coordinates))
})

test_that("precision_thinning handles different precisions", {
  # Precision = 4
  result_4 <- precision_thinning(coordinates, precision = 3, priority = priority)
  expect_equal(sum(result_4[[1]]), 2)

  # Precision = 2
  result_3 <- precision_thinning(coordinates, precision = 2)
  expect_equal(sum(result_3[[1]]), 1)
})

test_that("precision_thinning returns multiple trials correctly", {
  result <- precision_thinning(coordinates, trials = 5, all_trials = TRUE)

  expect_equal(length(result), 5)
  expect_true(all(sapply(result, is.logical)))
})

test_that("precision_thinning handles identical coordinates", {
  coordinates <- matrix(c(-123.3656, -123.3656, -123.3656, 48.4284, 48.4284, 48.4284), ncol = 2)
  result <- precision_thinning(coordinates, precision = 4)

  expect_equal(sum(result[[1]]), 1)
})

test_that("precision_thinning provides consistent results with fixed seed", {
  set.seed(42)
  result1 <- precision_thinning(coordinates, trials = 1, all_trials = FALSE)

  set.seed(42)
  result2 <- precision_thinning(coordinates, trials = 1, all_trials = FALSE)

  expect_equal(result1, result2)
})

test_that("precision_thinning handles a single coordinate", {
  coordinates <- matrix(c(-123.3656, 48.4284), ncol = 2)
  result <- precision_thinning(coordinates)

  expect_equal(length(result[[1]]), 1)
  expect_true(result[[1]])
})

test_that("precision_thinning validates input parameters", {
  expect_error(precision_thinning(coordinates, precision = -1), "`precision` must be a non-negative integer.")
  expect_error(precision_thinning(coordinates, precision = 2, priority = 1), "'priority' must be a numeric vector with same length as number of points.")
  expect_warning(precision_thinning(coordinates, precision = 2, priority = c(NA, 2, 3)), "NA values found in 'priority'.")
})

#' Check for Longitude/Latitude Coordinates
#'
#' This function checks whether a pair of coordinate vectors represent
#' geographic longitude and latitude values. The function returns `TRUE`
#' if all longitude values are within -180 (- tolerance) and 180 (+ tolerance)
#' and the latitude is within -90 (- tolerance) and 90 (+ tolerance).
#'
#' @param lon Numeric vector of longitudes in degrees.
#' @param lat Numeric vector of latitudes in degrees.
#' @param tolerance Numeric tolerance (in degrees) for checking the global
#'   longitude/latitude bounds. Default is 0.1.
#'
#' @return A logical value. `TRUE` if values are within the ranges and `FALSE` otherwise.
#'
#' @examples
#' is_lonlat(lon = c(-3, 10, 179), lat = c(40, -20, 5))
#' is_lonlat(lon = c(100000, 150000), lat = c(4500000, 4600000))
#'
#' @export
is_lonlat <- function(lon, lat, tolerance = 0.1) {
  lon_ok <- all(lon >= -180 - tolerance & lon <= 180 + tolerance, na.rm = TRUE)
  lat_ok <- all(lat >= -90  - tolerance & lat <=  90 + tolerance, na.rm = TRUE)

  return(lon_ok && lat_ok)
}

#' Convert Geographic Coordinates to Cartesian Coordinates
#'
#' This function converts geographic coordinates, given as longitude and latitude in degrees, to Cartesian coordinates (x, y, z) assuming a spherical Earth model.
#'
#' @param lon Numeric vector of longitudes in degrees.
#' @param lat Numeric vector of latitudes in degrees.
#' @param R Radius of the Earth in kilometers (default: 6371 km).
#' @return A numeric matrix with three columns (x, y, z) representing Cartesian coordinates.
#' @examples
#' lon <- c(-122.4194, 0)
#' lat <- c(37.7749, 0)
#' lon_lat_to_cartesian(lon, lat)
#' @export
lon_lat_to_cartesian <- function(lon, lat, R = 6371) {
  lat_rad <- lat * pi / 180
  lon_rad <- lon * pi / 180
  x <- R * cos(lat_rad) * cos(lon_rad)
  y <- R * cos(lat_rad) * sin(lon_rad)
  z <- R * sin(lat_rad)
  return(cbind(x, y, z))
}

#' Estimate Maximum Neighbors for kd-Tree Thinning
#'
#' This function estimates the maximum value of k (the number of nearest neighbors)
#' for kd-tree-based thinning by evaluating the densest regions of a spatial dataset.
#' The function uses a histogram-based binning approach for efficiency and low memory usage.
#'
#' @param coordinates A matrix of spatial coordinates with two columns for longitude and latitude.
#' @param thin_dist A positive numeric value representing the thinning distance in kilometers.
#'        This defines the resolution of the grid used for density calculations.
#' @param distance Distance metric used `c("haversine", "euclidean")`.
#'
#' @return A numeric value representing the maximum k (number of nearest neighbors) required
#'         for the densest regions in the dataset.
#'
#' @details
#' The function divides the spatial domain into grid cells based on the specified thinning distance.
#' Grid cell sizes are determined assuming approximately 111.32 km per degree (latitude/longitude).
#' The function identifies the densest grid cells and their immediate neighbors to compute the maximum k value.
#'
#' @examples
#' # Generate sample data
#' set.seed(123)
#' coordinates <- matrix(runif(200, min = -10, max = 10), ncol = 2)
#'
#' # Estimate k for kd-tree thinning
#' k_max <- estimate_k_max(coordinates, thin_dist = 50)
#' print(k_max)
#'
#' @export
estimate_k_max <- function(coordinates, thin_dist, distance = c("haversine", "euclidean")) {
  # Validate inputs
  if (!is.numeric(thin_dist) || thin_dist <= 0) {
    stop("`thin_dist` must be a positive number.")
  }
  if (!is.matrix(coordinates) || ncol(coordinates) != 2) {
    stop("`coordinates` must be a matrix with two columns (longitude and latitude).")
  }
  distance <- match.arg(distance)

  if (distance == "haversine") {
    # Convert thinning distance to degrees (approximation for latitude/longitude)
    cell_size <- thin_dist / 111.32
  } else {
    cell_size <- thin_dist
  }

  # Create grid breaks based on cell size
  x_breaks <- seq(min(coordinates[, 1]), max(coordinates[, 1]), by = cell_size)
  y_breaks <- seq(min(coordinates[, 2]), max(coordinates[, 2]), by = cell_size)
  x_breaks[length(x_breaks)] <- max(coordinates[, 1])
  y_breaks[length(y_breaks)] <- max(coordinates[, 2])

  if (length(x_breaks) == 1){
    x_breaks <- c(min(coordinates[, 1]), max(coordinates[, 1]))
  }
  if (length(y_breaks) == 1){
    y_breaks <- c(min(coordinates[, 2]), max(coordinates[, 2]))
  }

  # Bin and count the data
  bin_counts <- table(
    cut(coordinates[, 1], breaks = x_breaks, include.lowest = TRUE),
    cut(coordinates[, 2], breaks = y_breaks, include.lowest = TRUE)
  )
  bin_matrix <- as.matrix(bin_counts)

  # Identify the densest grid cells
  max_density_cells <- which(bin_matrix >= stats::quantile(bin_matrix, probs = 0.95), arr.ind = TRUE)

  # Calculate maximum k (including neighbors)
  max_k <- max(apply(max_density_cells, 1, function(cell) {
    # Extract the row and column of the current cell
    row <- cell[1]
    col <- cell[2]

    # Sum counts in the current cell and its neighbors
    sum(bin_matrix[
      max(1, row - 1):min(nrow(bin_matrix), row + 1),
      max(1, col - 1):min(ncol(bin_matrix), col + 1)
    ])
  }))

  return(max_k)
}

#' Compute Nearest Neighbor Distances
#'
#' Calculates nearest neighbor distances using geodesic or Euclidean distance.
#'
#' @param coordinates A matrix of coordinates with two columns.
#' @param distance A character string: "haversine" (default) or "euclidean".
#' @param R Radius of the Earth in kilometers. Default is 6371.
#'
#' @return A numeric vector of nearest neighbor distances, in meters (haversine) or in map units (euclidean).
#'
#' @examples
#' # Example with geographic (longitude/latitude) coordinates
#' set.seed(123)
#' coords_geo <- matrix(cbind(runif(10, -10, 10), runif(10, 40, 50)), ncol = 2)
#' nnd_haversine <- compute_nearest_neighbor_distances(coords_geo, distance = "haversine")
#' print(round(nnd_haversine, 2))  # in km
#'
#' # Example with projected coordinates (Euclidean)
#' coords_proj <- matrix(runif(20), ncol = 2) * 100  # e.g., meters or map units
#' nnd_euclidean <- compute_nearest_neighbor_distances(coords_proj, distance = "euclidean")
#' print(round(nnd_euclidean, 2))
#'
#' @export
compute_nearest_neighbor_distances <- function(coordinates, distance = "haversine", R = 6371) {
  if (nrow(coordinates) < 2) return(rep(NA, nrow(coordinates)))

  kd_result <- nabor::knn(coordinates, k = 2)
  nearest_idx <- kd_result$nn.idx[, 2]

  if (distance == "haversine") {
    dists <- numeric(nrow(coordinates))
    for (i in seq_len(nrow(coordinates))) {
      dists[i] <- fields::rdist.earth(coordinates[i, , drop = FALSE],
                                      coordinates[nearest_idx[i], , drop = FALSE],
                                      miles = FALSE, R = R)[1, 1]  # km
    }
  } else {
    # Return raw Euclidean distance from kd-tree
    dists <- kd_result$nn.dists[, 2]  # already in native units
  }

  return(dists)
}


#' Calculate Spatial Coverage (Convex Hull Area)
#'
#' Computes the area of the convex hull formed by the points.
#' Uses geodetic area (km2) if coordinates are lon/lat and distance = "haversine",
#' otherwise computes area in squared map units.
#'
#' @param coordinates A matrix of coordinates (longitude and latitude or planar x/y).
#' @param distance A character string: "haversine" (default) or "euclidean".
#'
#' @return A numeric value representing the convex hull area (km2 or unit2).
#'
#' @examples
#' # Geographic coordinates (lon/lat)
#' set.seed(456)
#' coords_geo <- matrix(cbind(runif(10, -10, 10), runif(10, 40, 50)), ncol = 2)
#' area_haversine <- calculate_spatial_coverage(coords_geo, distance = "haversine")
#' print(round(area_haversine, 2))  # in km2
#'
#' # Projected coordinates (Euclidean/map units)
#' coords_proj <- matrix(runif(20), ncol = 2) * 100  # e.g., map units
#' area_euclidean <- calculate_spatial_coverage(coords_proj, distance = "euclidean")
#' print(round(area_euclidean, 2))  # in unit2
#'
#' @export
calculate_spatial_coverage <- function(coordinates, distance = "haversine") {
  if (nrow(unique(coordinates)) < 3) return(0)

  if (distance == "haversine") {
    sf_pts <- sf::st_as_sf(data.frame(coordinates), coords = c(1, 2), crs = 4326)
    hull <- sf::st_convex_hull(sf::st_union(sf_pts))
    area <- as.numeric(sf::st_area(hull)) / 1e6 # km2
  } else {
    # Simple polygon area in planar units
    ch <- grDevices::chull(coordinates)
    hull_coords <- coordinates[c(ch, ch[1]), ]
    area <- abs(sum(hull_coords[-1, 1] * hull_coords[-nrow(hull_coords), 2] -
                      hull_coords[-nrow(hull_coords), 1] * hull_coords[-1, 2])) / 2
  }

  return(area)
}

coords <- data.frame(
  lon = c(-122.4194, -122.4195, -122.4196, -122.4197),
  lat = c(37.7749, 37.7740, 37.7741, 37.7750)
)
coords <- as.matrix(coords)

test_that("thin_points handles invalid input", {
  expect_error(estimate_k_max (NULL, 3), "`coordinates` must be a matrix")
  expect_error(estimate_k_max(coords, thin_dist = -2), "`thin_dist` must be a positive number.")
})

test_that("calculate_spatial_coverage with Euclidean distance", {
  coords_proj <- matrix(runif(20), ncol = 2) * 100
  area_euclidean <- calculate_spatial_coverage(coords_proj, distance = "euclidean")
  expect_true(area_euclidean > 0)
})

#' GeoThinned Object Constructor and Methods
#'
#' Create and interact with spatial thinning results stored in a GeoThinned object.
#'
#' @param retained A list of logical vectors indicating retained points per trial.
#' @param method The thinning method used (e.g., "distance", "grid", "precision").
#' @param params A list of parameters used in thinning.
#' @param original_data The original unmodified data.
#' @param x An object of class \code{GeoThinned}.
#' @param object An object of class \code{GeoThinned}.
#' @param trial Integer index of the thinning trial to extract (for \code{summary()}, \code{get_trial()}, \code{plot()}, \code{as_sf()}). Default `NULL`, which will return the largest dataset.
#' @param crs Coordinate reference system to assign to the resulting \code{sf} object (optional).
#' @param show_original Logical, whether to show original points.
#' @param col_original Colors for original points.
#' @param col_thinned Colors for thinned points.
#' @param pch_original Point shapes for original points.
#' @param pch_thinned Point shapes for thinned points.
#' @param main Title of the plot.
#' @param ... Additional arguments (ignored).
#'
#' @details
#' When the 's2' package is not installed or the installed version is lower than 1.1.0,
#' the spatial coverage metric for geographic coordinates from the \code{summary()} is not computed.
#' This is an optional summary metric and does not affect the thinning process or any other package functionality.
#'
#' @return A \code{GeoThinned} object or associated results (summary, plot, trial subset).
#' When `thin_points()` is run with `all_trials = FALSE`, the returned object contains only the largest trial; therefore all methods refer to this single subset.
#' @seealso \code{\link{thin_points}}
#' @rdname GeoThinned
#' @export
as_GeoThinned <- function(retained, method, params = list(), original_data = NULL) {
  new_GeoThinned(retained = retained, method = method, params = params, original_data = original_data)
}

#' @rdname GeoThinned
#' @export
new_GeoThinned <- function(retained, method, params = list(), original_data = NULL) {
  structure(
    list(
      retained = retained,
      method = method,
      params = params,
      original_data = original_data
    ),
    class = "GeoThinned"
  )
}

#' @method print GeoThinned
#' @rdname GeoThinned
#' @export
print.GeoThinned <- function(x, ...) {
  cat("GeoThinned object\n")
  cat("Method used:", x$method, "\n")
  cat("Number of trials:", x$params$trials, paste0("(", length(x$retained), " trials returned)\n"))
  cat("Points retained in largest trial:", sum(x$retained[[largest_index(x)]]), "\n")
  invisible(x)
}

#' @method summary GeoThinned
#' @rdname GeoThinned
#' @export
summary.GeoThinned <- function(object, trial = NULL, ...) {
  if (is.null(trial)) {
    trial <- largest_index(object)
  }
  if (trial < 1 || trial > length(object$retained)) {
    stop("Invalid trial index. 'trial' must be between 1 and ", length(object$retained), ".")
  }

  kept <- object$retained[[trial]]
  lon_col <- object$params$lon_col
  lat_col <- object$params$lat_col
  distance_type <- if (!is.null(object$params$distance)) {object$params$distance} else {"haversine"}

  original_coords <- as.matrix(object$original_data[, c(lon_col, lat_col)])
  thinned_coords <- as.matrix(object$original_data[kept, c(lon_col, lat_col)])

  n_orig <- nrow(original_coords)
  n_thin <- sum(kept)
  point_retention <- n_thin / n_orig

  nnd_orig <- compute_nearest_neighbor_distances(original_coords, distance = distance_type)
  nnd_thin <- compute_nearest_neighbor_distances(thinned_coords, distance = distance_type)

  if (object$method %in% c("grid", "precision") && distance_type == "haversine") {
    message("Note: Nearest neighbor distances and spatial coverage are computed using Haversine geometry by default.")
  }

  if (n_thin < 3) {
    warning("Fewer than 3 points in thinned set. Spatial coverage may not be meaningful.")
  }

  has_updated_s2 <- requireNamespace("s2", quietly = TRUE) && utils::packageVersion("s2") >= "1.1.0"

  if (distance_type != "haversine" || has_updated_s2) {
    coverage_orig <- calculate_spatial_coverage(original_coords, distance = distance_type)
    coverage_thin <- calculate_spatial_coverage(thinned_coords, distance = distance_type)
  } else {
    message("Spatial coverage is an optional summary metric and is not computed because package 's2' (>= 1.1.0) is not available. Please install or update 's2' to enable this metric.")
    coverage_orig <- NA
    coverage_thin <- NA
  }

  summary <- list(
    trial = trial,
    method = object$method,
    distance_type = distance_type,
    n_points = list(original = n_orig, thinned = n_thin),
    retention_ratio = point_retention,
    nnd_summary = list(
      original = summary(nnd_orig),
      thinned = summary(nnd_thin)
    ),
    spatial_coverage = c(original = coverage_orig, thinned = coverage_thin)
  )

  class(summary) <- "summary.GeoThinned"
  print(summary)
  invisible(summary)
}

#' @method print summary.GeoThinned
#' @rdname GeoThinned
#' @export
print.summary.GeoThinned <- function(x, ...) {
  nnd_units <- if (x$distance_type == "haversine") {"km"} else {"map units"}
  area_units <- if (x$distance_type == "haversine") {"km2"} else {"map units2"}

  cat("Summary of GeoThinneR Results\n")
  cat("-----------------------------\n")
  cat("Trial summarized  :", x$trial, "\n")
  cat("Method used       :", x$method, "\n")
  cat("Distance metric   :", x$distance_type, "\n\n")

  cat("Number of points:\n")
  cat(sprintf("  %-15s %10d\n", "Original", x$n_points$original))
  cat(sprintf("  %-15s %10d\n", "Thinned", x$n_points$thinned))
  cat(sprintf("  %-15s %10.3f\n\n", "Retention ratio", x$retention_ratio))

  cat("Nearest Neighbor Distances [", nnd_units, "]\n", sep = "")
  nnd_tbl <- rbind(
    Original = unname(round(x$nnd_summary$original[c("Min.", "1st Qu.", "Median", "Mean", "3rd Qu.", "Max.")], 3)),
    Thinned = unname(round(x$nnd_summary$thinned[c("Min.", "1st Qu.", "Median", "Mean", "3rd Qu.", "Max.")], 3))
  )
  colnames(nnd_tbl) <- c("Min", "1st Qu.", "Median", "Mean", "3rd Qu.", "Max")
  print(nnd_tbl, right = TRUE)
  cat("\n")

  cat("Spatial Coverage (Convex Hull Area) [", area_units, "]\n", sep = "")
  cat(sprintf("  %-15s %10.3f\n", "Original", x$spatial_coverage["original"]))
  cat(sprintf("  %-15s %10.3f\n", "Thinned", x$spatial_coverage["thinned"]))
}

#' @method plot GeoThinned
#' @rdname GeoThinned
#' @importFrom graphics points legend
#' @export
plot.GeoThinned <- function(x, trial = NULL,
                            show_original = TRUE,
                            col_original = "#EB714B",
                            col_thinned = "#5183B3",
                            pch_original = 1,
                            pch_thinned = 16,
                            main = NULL,
                            ...) {
  # Validate trial index
  if (is.null(trial)) {
    trial <- largest_index(x)
  }
  if (trial < 1 || trial > length(x$retained)) {
    stop("Invalid trial index. 'trial' must be between 1 and ", length(x$retained), ".")
  }

  kept <- x$retained[[trial]]
  if (!is.logical(kept) || length(kept) != nrow(x$original_data)) {
    stop("Invalid retained vector: must be logical and match original data length.")
  }

  lon_col <- x$params$lon_col
  lat_col <- x$params$lat_col
  coords <- x$original_data[, c(lon_col, lat_col)]

  if (is.null(main)) {
    main <- paste("GeoThinneR - Trial", trial)
  }

  if (show_original) {
    plot(coords, col = col_original, pch = pch_original,
         xlab = "Longitude", ylab = "Latitude",
         main = main, ...)
    points(coords[kept, ], col = col_thinned, pch = pch_thinned)
    legend("topleft",
           legend = c("Removed", "Retained"),
           col = c(col_original, col_thinned),
           pch = c(pch_original, pch_thinned),
           bty = "o", bg="#faf5ef")
  } else {
    plot(coords[kept, ], col = col_thinned, pch = pch_thinned,
         xlab = "Longitude", ylab = "Latitude",
         main = main, ...)
  }
}

#' @rdname GeoThinned
#' @export
largest <- function(x, ...) {
  UseMethod("largest")
}

#' @method largest GeoThinned
#' @rdname GeoThinned
#' @export
largest.GeoThinned <- function(x, ...) {
  x$original_data[x$retained[[largest_index(x)]], , drop = FALSE]
}

#' @rdname GeoThinned
#' @export
largest_index <- function(x, ...) {
  UseMethod("largest_index")
}

#' @method largest_index GeoThinned
#' @rdname GeoThinned
#' @export
largest_index.GeoThinned <- function(x, ...) {
  sizes <- sapply(x$retained, sum)
  which.max(sizes)
}

#' @rdname GeoThinned
#' @export
get_trial <- function(x, trial = NULL, ...) {
  UseMethod("get_trial")
}

#' @method get_trial GeoThinned
#' @rdname GeoThinned
#' @export
get_trial.GeoThinned <- function(x, trial = NULL, ...) {
  if (is.null(trial)) {
    trial <- largest_index(x)
  }
  x$original_data[x$retained[[trial]], , drop = FALSE]
}

#' @rdname GeoThinned
#' @export
as_sf <- function(x, ...) {
  UseMethod("as_sf")
}

#' @method as_sf GeoThinned
#' @rdname GeoThinned
#' @export
as_sf.GeoThinned <- function(x, trial = NULL, crs = 4326, ...) {
  if (is.null(trial)) {
    trial <- largest_index(x)
  }

  retained_ids <- x$retained[[trial]]
  retained_points <- x$original_data[retained_ids, , drop = FALSE]

  # Extract lon/lat columns
  lon_col <- x$params$lon_col
  lat_col <- x$params$lat_col
  if (is.null(lon_col) || is.null(lat_col)) {
    stop("Could not determine longitude and latitude columns.")
  }

  # Convert to sf
  sf::st_as_sf(retained_points, coords = c(lon_col, lat_col), crs = crs, ...)
}

# Generate sample data
set.seed(123)
sample_data <- data.frame(
  lon = runif(100, -180, 180),
  lat = runif(100, -90, 90)
)

test_that("thin_points works with valid input", {
  # Test thinning using default method (distance-based thinning using kd-tree method)
  thinned_data <- thin_points(sample_data,
                              lon_col = "lon",
                              lat_col = "lat",
                              trials = 5,
                              seed = 123,
                              verbose = TRUE)
  expect_equal(length(thinned_data$retained[[1]]), 100) # Check the number of rows in the first trial
})

test_that("thin_points works with tibble and matrix input", {
  skip_if_not_installed("tibble")
  sample_tibble <- tibble::as_tibble(sample_data)

  res_tibble <- thin_points(sample_tibble, trials = 2)
  expect_s3_class(res_tibble, "GeoThinned")
  trial_data <- get_trial(res_tibble, 1)
  expect_s3_class(trial_data, "data.frame")
  expect_true(all(c("lon", "lat") %in% names(trial_data)))

  sample_matrix <- as.matrix(sample_data[, 1:2])
  colnames(sample_matrix) <- NULL
  res_matrix <- thin_points(sample_matrix, lon_col = NULL, lat_col = NULL, trials = 2)
  expect_s3_class(res_matrix, "GeoThinned")
  trial_data <- get_trial(res_matrix, 1)
  expect_s3_class(trial_data, "data.frame")
  expect_equal(ncol(trial_data), 2)
})

test_that("thin_points works with distance thinning", {
  res <- thin_points(sample_data, method = "distance", thin_dist = 10, trials = 2)
  expect_s3_class(res, "GeoThinned")
  expect_equal(length(res$retained), 1)
  expect_equal(nrow(get_trial(res, 1)), sum(res$retained[[1]]))
})

test_that("thin_points works with grid thinning", {
  res <- thin_points(sample_data, method = "grid", thin_dist = 10, trials = 2)
  expect_s3_class(res, "GeoThinned")
  expect_equal(length(res$retained), 1)
  expect_equal(nrow(get_trial(res, 1)), sum(res$retained[[1]]))
})

test_that("thin_points works with precision thinning", {
  res <- thin_points(sample_data, method = "precision", precision = 2, trials = 2)
  expect_s3_class(res, "GeoThinned")
  expect_equal(length(res$retained), 1)
  expect_equal(nrow(get_trial(res, 1)), sum(res$retained[[1]]))
})

test_that("thin_points works with grouped data", {
  sample_data$species <- sample(c("species_A", "species_B"), 100, replace = TRUE)

  res <- thin_points(sample_data,
                     group_col = "species",
                     method = "distance",
                     thin_dist = 1000,
                     trials = 2)

  expect_s3_class(res, "GeoThinned")
  expect_equal(nrow(res$original_data), 100)
  expect_equal(length(res$retained), 1)
  retained_total <- sum(res$retained[[1]])
  expect_true(retained_total < 100)  # actual thinning occurred
})

test_that("largest() returns correct trial from GeoThinned object", {
  res <- thin_points(sample_data, method = "distance", trials = 1)
  b <- largest(res)
  expect_equal(sum(res$retained[[1]]), nrow(b))
})

test_that("thin_points handles invalid inputs correctly", {
  expect_error(thin_points(NULL), "Specified longitude or latitude columns do not exist")
  expect_error(thin_points(data.frame(x = 1:10, y = 1:10), lon_col = "lon", lat_col = "lat"),
               "Specified longitude or latitude columns do not exist")
  expect_error(thin_points(data.frame(lon = c("a", "b", "c"),lat = 1:3), lon_col = NULL), "Error: The first two columns must be numeric")
  expect_error(thin_points(sample_data, trials = -1), "`trials` must be a positive integer")
  expect_error(thin_points(sample_data, all_trials = "yes"), "`all_trials` must be a logical value")
  expect_error(thin_points(sample_data, group_col = "unknown"), "Specified grouping column does not exist")
  expect_warning(thin_points(data.frame(lon = c(10, 200), lat = c(20, 30))), "Input coordinates lie outside the typical longitude/latitude global ranges.")
})

test_that("GeoThinned object constructor works", {
  data <- data.frame(
    lon = runif(10, -10, 10),
    lat = runif(10, 40, 50)
  )
  retained <- list(sample(c(TRUE, FALSE), 10, replace = TRUE))
  method <- "distance"
  params <- list(lon_col = "lon", lat_col = "lat")

  obj <- as_GeoThinned(retained = retained, method = method, params = params, original_data = data)

  expect_s3_class(obj, "GeoThinned")
  expect_equal(obj$method, method)
  expect_equal(obj$original_data, data)
  expect_type(obj$retained, "list")
  expect_true(all(sapply(obj$retained, is.logical)))
})

test_that("print and summary work for GeoThinned", {
  data <- data.frame(
    lon = runif(10, -10, 10),
    lat = runif(10, 40, 50)
  )
  retained <- list(rep(TRUE, 10))
  obj <- as_GeoThinned(retained, method = "distance", params = list(lon_col = "lon", lat_col = "lat"), original_data = data)

  expect_output(print(obj), "GeoThinned object")
  expect_output(summary(obj), "Summary of GeoThinneR Results")
})

test_that("plot.GeoThinned runs without error", {
  data <- data.frame(
    lon = runif(10, -10, 10),
    lat = runif(10, 40, 50)
  )
  retained <- list(rep(TRUE, 10))
  obj <- as_GeoThinned(retained, method = "distance", params = list(lon_col = "lon", lat_col = "lat"), original_data = data)

  expect_silent(plot(obj))
  expect_silent(plot(obj, show_original = FALSE))
})

test_that("get_trial and largest work", {
  data <- data.frame(
    lon = runif(10, -10, 10),
    lat = runif(10, 40, 50)
  )
  retained <- list(sample(c(TRUE, FALSE), 10, replace = TRUE),
                   sample(c(TRUE, FALSE), 10, replace = TRUE))
  obj <- as_GeoThinned(retained, method = "distance", params = list(lon_col = "lon", lat_col = "lat"), original_data = data)

  expect_true(is.data.frame(get_trial(obj, 1)))
  expect_true(is.data.frame(largest(obj)))
  expect_type(largest_index(obj), "integer")
  expect_gt(largest_index(obj), 0)
})

test_that("as_sf.GeoThinned returns an sf object", {
  data <- data.frame(
    lon = runif(10, -10, 10),
    lat = runif(10, 40, 50)
  )
  retained <- list(rep(TRUE, 10))
  obj <- as_GeoThinned(retained, method = "distance", params = list(lon_col = "lon", lat_col = "lat"), original_data = data)

  result <- as_sf(obj)
  expect_s3_class(result, "sf")
  expect_true("geometry" %in% colnames(result))
})

#' Thinning Algorithm for Spatial Data
#'
#' This function performs the core thinning algorithm used to reduce the density of points
#' in spatial data while maintaining spatial representation. It iteratively removes the
#' points with the most neighbors until no points with neighbors remain. The algorithm
#' supports multiple trials to find the optimal thinning solution.
#'
#' @param neighbor_indices A list of integer vectors where each element contains the indices of the neighboring points for each point in the dataset.
#' @param trials A positive integer specifying the number of thinning trials to perform. Default is 10.
#' @param all_trials A logical value indicating whether to return results of all attempts (`TRUE`) or only the best attempt with the most points retained (`FALSE`). Default is `FALSE`.
#' @param priority A numeric vector of the same length as the number of points, specifying a priority weight for each point. Higher values indicate higher importance and are favored when selecting which points to retain. Priority is used to guide selection when multiple candidate points are otherwise equally valid (e.g., points in the same grid cell, with the same rounded coordinates, or with the same number of neighbors).
#'
#' @return A list of logical vectors indicating which points are kept in each trial if all_trials is TRUE; otherwise, a list with a single logical vector indicating the points kept in the best trial.
#'
#' @examples
#' # Example usage within a larger thinning function
#' neighbor_indices <- list(c(2, 3), c(1, 3), c(1, 2))
#' trials <- 5
#' all_trials <- FALSE
#' kept_points <- max_thinning_algorithm(neighbor_indices, trials, all_trials)
#' print(kept_points)
#'
#' @export
max_thinning_algorithm <- function(neighbor_indices, trials, all_trials = FALSE, priority = NULL) {
  # Compute initial neighbor counts
  n <- length(neighbor_indices)
  neighbor_counts <- lengths(neighbor_indices)

  # Initialize results list for trials
  kept_points <- vector("list", ifelse(all_trials, trials, 1))
  kept_points[[1]] <- rep(FALSE, n)  # Start with no points kept

  # Main thinning loop for the specified number of trials
  for (i in seq_len(trials)) {
    kept_points_trial <- rep(TRUE, n)
    neighbor_counts_trial <- neighbor_counts
    max_neighbors <- max(neighbor_counts_trial)

    while (max_neighbors > 0) {  # Exit loop if no neighbors remain
      # Find indices of points with the maximum neighbors
      points_to_remove <- which(neighbor_counts_trial == max_neighbors)
      if (length(points_to_remove) > 1) {
        if (!is.null(priority)) {
          points_to_remove <- points_to_remove[sample(which(priority[points_to_remove] == min(priority[points_to_remove])), 1)]
        } else {
          points_to_remove <- points_to_remove[as.integer(ceiling(stats::runif(1, 0, length(points_to_remove))))]
        }
      }

      # Recompute neighbor counts for remaining points
      neighbor_counts_trial[neighbor_indices[[points_to_remove]]] <- neighbor_counts_trial[neighbor_indices[[points_to_remove]]] - 1
      neighbor_counts_trial[points_to_remove] <- 0L

      # Mark that point as not to keep
      kept_points_trial[points_to_remove] <- FALSE

      # Identify points to remove: those with the most neighbors
      max_neighbors <- max(neighbor_counts_trial)
    }

    # Store the results of this trial
    if (all_trials) {
      kept_points[[i]] <- kept_points_trial
    } else {
      # Update the kept_points if this trial resulted in more kept points
      if (sum(kept_points_trial) > sum(kept_points[[1]])) {
        kept_points[[1]] <- kept_points_trial
      }
    }
  }

  # Return list of trials or list with best trial in first position
  return(kept_points)
}

#' Select Target Number of Points for Spatial Thinning
#'
#' This function selects a specified number of points from a spatial dataset while maximizing
#' the distance between selected points.
#'
#' @param distance_matrix A matrix of pairwise distances between points.
#' @param target_points An integer specifying the number of points to retain.
#' @param thin_dist A positive numeric value representing the thinning distance in kilometers.
#' @param trials A positive integer specifying the number of thinning trials to perform. Default is 10.
#' @param all_trials A logical value indicating whether to return results of all attempts (`TRUE`) or only the best attempt with the most points retained (`FALSE`). Default is `FALSE`.
#'
#' @return A list of logical vectors indicating which points are kept in each trial if `all_trials` is `TRUE`; otherwise, a list with a single logical vector indicating the points kept in the best trial.
#'
#' @examples
#' # Example distance matrix (3 points)
#' dist_matrix <- matrix(c(0, 2, 5,
#'                         2, 0, 3,
#'                         5, 3, 0), ncol = 3)
#'
#' # Select 2 points maximizing distance
#'result <- select_target_points(dist_matrix, target_points = 2,
#'                               thin_dist = 4, trials = 5, all_trials = TRUE)
#'
#' @export
select_target_points <- function(distance_matrix, target_points, thin_dist, trials, all_trials = FALSE) {
  # Initialize results list for trials
  n <- nrow(distance_matrix)
  kept_points <- vector("list", ifelse(all_trials, trials, 1))
  kept_points[[1]] <- rep(FALSE, n) # Start with no points kept

  for (i in seq_len(trials)) {
    kept_points_trial <- rep(FALSE, n)
    points_to_keep <- as.integer(stats::runif(1, 1, n)) # Randomly select the first point

    while (length(points_to_keep) < target_points) {
      dist_to_closest_kept <- matrixStats::rowMins(distance_matrix[, points_to_keep, drop = FALSE], na.rm = TRUE)
      dist_to_closest_kept[points_to_keep] <- 0

      max_dist_closest <- max(dist_to_closest_kept, na.rm = TRUE)
      if (max_dist_closest < thin_dist) break  # Stop if no points are further than thin_dist

      furthest <- which(dist_to_closest_kept == max_dist_closest)
      if (length(furthest) > 1) {
        furthest <- sample(furthest, 1)  # Randomly select one if multiple furthest points
      }

      points_to_keep <- c(points_to_keep, furthest)
    }

    kept_points_trial[points_to_keep] <- TRUE

    if (all_trials) {
      kept_points[[i]] <- kept_points_trial
    } else {
      if (length(points_to_keep) == target_points) {
        kept_points[[1]] <- kept_points_trial
        break  # Exit if target points are reached
      } else if (sum(kept_points_trial) > sum(kept_points[[1]])) {
        kept_points[[1]] <- kept_points_trial  # Update if this trial kept more points
      }
    }
  }

  return(kept_points)
}

# Generate sample data
set.seed(123)
sample_data <- data.frame(
  lon = runif(100, 0, 1),
  lat = runif(100, 0, 1)
)

test_that("max_thinning retains maximum number of points", {
  neighbor_indices <- compute_neighbors_brute(as.matrix(sample_data), thin_dist = 10)
  result_max_thinning <- max_thinning_algorithm(neighbor_indices, trials = 10, all_trials = TRUE)

  expect_length(result_max_thinning[[1]], nrow(sample_data))
  expect_true(sum(result_max_thinning[[1]]) < nrow(sample_data)) # Should thin some points
})

test_that("select_target_points returns the exact number of requested points", {
  distance_matrix <- fields::RdistEarth(x1 = as.matrix(sample_data), miles = FALSE)
  result_target_points <- select_target_points(distance_matrix, target_points = 20, thin_dist = 10, trials = 10, all_trials = TRUE)
  expect_true(sum(result_target_points[[1]]) == 20) # Should thin some points

  result_target_points <- select_target_points(distance_matrix, target_points = 20, thin_dist = 10, trials = 10, all_trials = FALSE)
  expect_length(result_target_points[[1]], nrow(sample_data))

  distance_matrix <- fields::RdistEarth(x1 = as.matrix(rbind(sample_data, sample_data)), miles = FALSE)
  result_target_points <- select_target_points(distance_matrix, target_points = 20, thin_dist = 100, trials = 10, all_trials = FALSE)
  expect_length(result_target_points[[1]], 2*nrow(sample_data))
})

#' Perform Grid-Based Thinning of Spatial Points
#'
#' This function performs thinning of spatial points by assigning them to grid cells based on a specified resolution or thinning distance. It can either create a new raster grid or use an existing `terra::SpatRaster` object.
#'
#' @param coordinates A numeric matrix or data frame with two columns representing the x (longitude) and y (latitude) coordinates of the points.
#' @param thin_dist A numeric value representing the thinning distance in kilometers. It will be converted to degrees if `resolution` is not provided.
#' @param resolution A numeric value representing the resolution (in degrees) of the raster grid. If provided, this takes priority over `thin_dist`.
#' @param origin A numeric vector of length 2 (e.g., `c(0, 0)`), specifying the origin of the raster grid (optional).
#' @param raster_obj An optional `terra::SpatRaster` object to use for grid thinning. If provided, the raster object will be used instead of creating a new one.
#' @param n A positive integer specifying the maximum number of points to retain per grid cell (default: 1).
#' @param trials An integer specifying the number of trials to perform for thinning (default: 10).
#' @param all_trials A logical value indicating whether to return results for all trials (`TRUE`) or just the first trial (`FALSE`, default).
#' @param crs An optional CRS (Coordinate Reference System) to project the coordinates and raster (default WGS84, `epsg:4326`). This can be an EPSG code, a PROJ.4 string, or a `terra::crs` object.
#' @param priority A numeric vector of the same length as the number of points, specifying a priority weight for each point. Higher values indicate higher importance and are favored when selecting which points to retain. Priority is used to guide selection when multiple candidate points are otherwise equally valid (e.g., points in the same grid cell, with the same rounded coordinates, or with the same number of neighbors).
#' @return A list of logical vectors indicating which points to keep for each trial.
#' @examples
#' # Example: Grid thinning using thin_dist
#' coords <- matrix(c(-122.4194, 37.7749,
#'                         -122.4195, 37.7740,
#'                         -122.4196, 37.7741), ncol = 2, byrow = TRUE)
#'
#' result <- grid_thinning(coords, thin_dist = 10, trials = 5, all_trials = TRUE)
#' print(result)
#'
#' # Example: Grid thinning using a custom resolution
#' result_res <- grid_thinning(coords, resolution = 0.01, n = 2, trials = 5)
#' print(result_res)
#'
#' # Example: Using a custom raster object
#' library(terra)
#' rast_obj <- terra::rast(nrows = 100, ncols = 100, xmin = -123, xmax = -121, ymin = 36, ymax = 38)
#' result_raster <- grid_thinning(coords, raster_obj = rast_obj, trials = 5)
#' print(result_raster)
#'
#' @importFrom data.table := .N
#'
#' @export
grid_thinning <- function(coordinates, thin_dist = NULL, resolution = NULL, origin = NULL, raster_obj = NULL, n = 1,  trials = 10, all_trials = FALSE, crs = "epsg:4326", priority = NULL) {

  # Validate input
  if (is.null(thin_dist) && is.null(resolution) && is.null(raster_obj)) {
    stop("Either thin_dist, resolution, or raster_obj must be provided.")
  }

  if (!is.null(priority)){
    if (!is.numeric(priority) || length(priority) != nrow(coordinates)){
      stop("'priority' must be a numeric vector with same length as number of points.")
    }

    if (any(is.na(priority))){
      warning("NA values found in 'priority'. Replacing with lowest priority (-Inf).")
      priority[is.na(priority)] <- -Inf
    }
  }

  if (!is.numeric(n) || length(n) != 1 || n <= 0) {
    stop("`n` must be a positive integer specifying the maximum number of points per grid cell.")
  }

  # Calculate resolution if thin_dist is provided and resolution is NULL
  if (is.null(resolution) & !is.null(thin_dist)) {
    resolution <- thin_dist / 111.32  # Convert km to degrees
  }

  # Create or use provided raster
  if (is.null(raster_obj)) {
    r_grid <- terra::rast(
      xmin = min(coordinates[, 1]) - resolution,
      xmax = max(coordinates[, 1]) + resolution,
      ymin = min(coordinates[, 2]) - resolution,
      ymax = max(coordinates[, 2]) + resolution,
      resolution = resolution,
      crs = crs
    )
    if (!is.null(origin)) {
      terra::origin(r_grid) <- origin
    }
  } else {
    r_grid <- raster_obj
  }

  # Assign points to raster cells
  grid_cell <- terra::cellFromXY(r_grid, coordinates)
  #unique_grid_cell <- unique(stats::na.omit(grid_cell))

  # Initialize results list for trials
  keep_points <- vector("list", ifelse(all_trials, trials, 1))

  # Main thinning loop for the specified number of trials
  for (i in seq_along(keep_points)) {
    if (is.null(priority)){
      sort_order <- stats::runif(nrow(coordinates))
    } else {
      sort_order <- rank(-priority, ties.method = "random")
    }
    keep_points_trial <- data.table::data.table(
      id = seq_len(nrow(coordinates)),
      rand_order = sort_order,
      grid_cell = grid_cell
    )
    keep_points_trial <- keep_points_trial[order(keep_points_trial[["rand_order"]], decreasing = TRUE), ]

    # Rank points within each grid cell and retain up to n points
    keep_points_trial[, "rank_within_cell" := seq_len(.N), by = "grid_cell"]
    keep_points_trial[, "keep"] <- keep_points_trial[, "rank_within_cell"] <= n
    #keep_points_trial[, "keep"] <- !duplicated(keep_points_trial[["grid_cell"]])

    # If not saving all trials, keep only the first trial result
    keep_points[[i]] <- keep_points_trial[order(keep_points_trial[["id"]]),][["keep"]]
  }

  return(keep_points)
}

coordinates <- matrix(c(-122.4194, 37.7749,
                        -122.4195, 37.7740,
                        -122.4196, 37.7741), ncol = 2, byrow = TRUE)
priority <- 1:3


test_that("grid_thinning works with thin_dist", {
  result <- grid_thinning(coordinates, thin_dist = 10, trials = 3, priority = priority)

  expect_length(result[[1]], nrow(coordinates))
  expect_true(sum(result[[1]]) < nrow(coordinates)) # Should thin some points
})

test_that("grid_thinning works with resolution", {
  result_res <- grid_thinning(coordinates, resolution = 0.01, trials = 3, origin = c(0, 0))

  expect_length(result_res[[1]], nrow(coordinates))
  expect_true(sum(result_res[[1]]) < nrow(coordinates)) # Should thin some points
})

test_that("grid_thinning works with a raster object", {
  library(terra)
  rast_obj <- terra::rast(nrows = 100, ncols = 100, xmin = -123, xmax = -121, ymin = 36, ymax = 38)

  result_raster <- grid_thinning(coordinates, raster_obj = rast_obj, trials = 3)

  expect_length(result_raster[[1]], nrow(coordinates))
  expect_true(sum(result_raster[[1]]) < nrow(coordinates)) # Should thin some points
})

test_that("grid_thinning returns multiple trials when all_trials = TRUE", {
  result <- grid_thinning(coordinates, thin_dist = 10, trials = 3, all_trials = TRUE)

  expect_length(result, 3)
  expect_true(all(sapply(result, length) == nrow(coordinates)))
})

test_that("grid_thinning works with priority", {
  result_priority <- grid_thinning(coordinates, thin_dist = 10, trials = 3, priority = priority)

  expect_length(result_priority[[1]], nrow(coordinates))
  expect_true(sum(result_priority[[1]]) < nrow(coordinates)) # Should thin some points
})

test_that("grid_thinning invalid input", {
  # grid_thinning without raster information
  expect_error(grid_thinning(coordinates), "Either thin_dist, resolution, or raster_obj must be provided.")

  # grid_thinning with wrongly formated priority
  expect_error(grid_thinning(coordinates, resolution = 2, priority = 1), "'priority' must be a numeric vector with same length as number of points.")

  # grid_thinning with wrongly formated n
  expect_error(grid_thinning(coordinates, resolution = 2, n = -1), "`n` must be a positive integer specifying the maximum number of points per grid cell.")

  # grid_thinning with NA priority values warning
  expect_warning(grid_thinning(coordinates, resolution = 2, priority = c(NA, 2, 3)), "NA values found in 'priority'.")


})

#' Perform Distance-Based Thinning
#'
#' This function applies a distance-based thinning algorithm using a kd-tree or brute-force approach.
#' Two modified algorithms based on kd-trees (local kd-trees and estimating the maximum number of neighbors) are implemented which scale better for large datasets.
#' The function removes points that are closer than a specified distance to each other while maximizing spatial representation.
#'
#' @param coordinates A matrix of coordinates to thin, with two columns representing longitude and latitude.
#' @param thin_dist A positive numeric value representing the thinning distance in kilometers.
#' @param trials An integer specifying the number of trials to run for thinning. Default is 10.
#' @param all_trials A logical indicating whether to return results of all attempts (`TRUE`) or only the best attempt with the most points retained (`FALSE`). Default is `FALSE`.
#' @param search_type A character string indicating the neighbor search method `c("local_kd_tree", "k_estimation", "kd_tree", "brute")`. The default value is `local_kd_tree`. See details.
#' @param target_points Optional integer specifying the number of points to retain. If `NULL` (default), the function tries to maximize the number of points retained.
#' @param priority A numeric vector of the same length as the number of points, specifying a priority weight for each point. Higher values indicate higher importance and are favored when selecting which points to retain. Priority is used to guide selection when multiple candidate points are otherwise equally valid (e.g., points in the same grid cell, with the same rounded coordinates, or with the same number of neighbors).
#' @param distance Distance metric to use `c("haversine", "euclidean")`. Default is Haversine for geographic coordinates.
#' @param R Radius of the Earth in kilometers (default: 6371 km).
#' @param n_cores Number of cores for parallel processing (only for `"local_kd_tree"`). Default is 1.
#'
#' @details
#' - `"kd_tree"`: Uses a single kd-tree for efficient nearest-neighbor searches.
#' - `"local_kd_tree"`: Builds multiple smaller kd-trees for better scalability.
#' - `"k_estimation"`: Approximates a maximum number of neighbors per point to reduce search complexity.
#' - `"brute"`: Computes all pairwise distances (inefficient for large datasets).
#'
#' @return A list. If `all_trials` is `FALSE`, the list contains a single logical vector indicating which points are kept in the best trial. If `all_trials` is `TRUE`, the list contains a logical vector for each trial.
#'
#' @examples
#' # Generate sample coordinates
#' set.seed(123)
#' result  <- matrix(runif(20, min = -180, max = 180), ncol = 2) # 10 random points
#'
#' # Perform thinning with local kd-trees
#' result_partitioned <- distance_thinning(result , thin_dist = 5000, trials = 5,
#'                                        search_type = "local_kd_tree", all_trials = TRUE)
#' print(result_partitioned)
#'
#' # Perform thinning estimating max number of neighbors
#' result_estimated <- distance_thinning(result , thin_dist = 5000, trials = 5,
#'                                        search_type = "k_estimation", all_trials = TRUE)
#' print(result_estimated)
#'
#' @export
distance_thinning <- function(coordinates, thin_dist = 10, trials = 10, all_trials = FALSE,
                              search_type = c("local_kd_tree", "k_estimation", "kd_tree", "brute"),
                              target_points = NULL, priority = NULL, distance = c("haversine", "euclidean"),
                              R = 6371, n_cores = 1) {

  # Input validation
  if (!is.matrix(coordinates) || ncol(coordinates) != 2) {
    stop("`coordinates` must be a matrix with two columns (longitude and latitude).")
  }
  if (!is.numeric(thin_dist) || thin_dist <= 0) {
    stop("`thin_dist` must be a positive number.")
  }
  distance <- match.arg(distance)
  search_type <- match.arg(search_type)
  if (!is.null(target_points)){ # Use brute-force algorithm as we need all pairwise distances
    if (!is.numeric(target_points) || target_points <= 0) {
      stop("`target_points` must be a positive number.")
    }
    message("For specific target points, brute force method is used.")
    search_type <- "brute"
  }

  if (!is.null(priority)){
    if (!is.numeric(priority) | length(priority) != nrow(coordinates)){
      stop("'priority' must be a numeric vector with same length as number of points.")
    }

    if (any(is.na(priority))){
      warning("NA values found in 'priority'. Replacing with lowest priority (-Inf).")
      priority[is.na(priority)] <- -Inf
    }
  }

  if (is.null(target_points)){
    # Determine neighbor search method
    neighbor_indices <- switch(
      search_type,
      "brute" = compute_neighbors_brute(coordinates, thin_dist, distance, R),
      "kd_tree" = compute_neighbors_kdtree(coordinates, thin_dist, nrow(coordinates), distance, R),
      "local_kd_tree" = compute_neighbors_local_kdtree(coordinates, thin_dist, distance, R, n_cores),
      "k_estimation" = compute_neighbors_kdtree(coordinates, thin_dist, estimate_k_max(coordinates, thin_dist, distance), distance, R),
      stop("Unsupported `search_type`. Choose from 'kd_tree', 'local_kd_tree', 'k_estimation', or 'brute'.")
    )

    # Run thinning algorithm to keep as max points as possible
    kept_points <- max_thinning_algorithm(neighbor_indices, trials, all_trials = all_trials, priority = priority)
  } else { # Try to select the exact number of points as much separated as possible
    # Compute neighbors
    if (distance == "haversine") {
      dist_mat <- fields::rdist.earth(x1 = coordinates, miles = FALSE, R = R)
    } else {
      dist_mat <- as.matrix(stats::dist(coordinates, upper = TRUE))
    }
    diag(dist_mat) <- NA  # Avoid self-matches

    kept_points <- select_target_points(dist_mat, target_points, thin_dist, trials, all_trials)
  }

  return(kept_points)
}

#' Compute Neighbors Using Brute-Force
#'
#' Computes neighbors for each point in a set of coordinates using a greedy
#' approach. All pairwise distances are calculated to identify neighbors within
#' a specified distance threshold.
#'
#' @param coordinates A matrix of coordinates to thin, with two columns representing longitude and latitude.
#' @param thin_dist A positive numeric value representing the thinning distance in kilometers.
#' @param distance A character string specifying the distance metric to use `c("haversine", "euclidean")`.
#' @param R A numeric value representing the radius of the Earth in kilometers. The default is 6371 km.
#'
#' @return A list where each element corresponds to a point and contains the indices of its neighbors.
#'
#' @examples
#' set.seed(123)
#' coords <- matrix(runif(20, min = -180, max = 180), ncol = 2)
#'
#' # Compute neighbors using brute fore
#' neighbors <- compute_neighbors_brute(coords, thin_dist = 10,)
#'
#' @export
compute_neighbors_brute <- function(coordinates, thin_dist, distance = c("haversine", "euclidean"), R = 6371) {
  # Initialize a list to store neighbor indices
  n <- nrow(coordinates)
  neighbor_indices <- vector("list", n)
  distance <- match.arg(distance)

  # Compute neighbors
  if (distance == "haversine") {
    dist_mat <- fields::rdist.earth(x1 = coordinates, miles = FALSE, R = R) < thin_dist
  } else {
    dist_mat <- as.matrix(stats::dist(coordinates, upper = TRUE)) < thin_dist
  }
  diag(dist_mat) <- NA  # Avoid self-matches
  for (i in seq_len(n)) {
    neighbor_indices[[i]] <- which(dist_mat[i, ])
  }

  return(neighbor_indices)
}

#' Compute Neighbors Using kd-Tree
#'
#' Computes neighbors for each point in a set of coordinates using a kd-tree
#' for efficient neighbor searches. This method is particularly useful for large datasets.
#'
#' @param coordinates A matrix of coordinates to thin, with two columns representing longitude and latitude.
#' @param thin_dist A positive numeric value representing the thinning distance in kilometers.
#' @param k An integer specifying the maximum number of neighbors to consider for each point.
#' @param distance A character string specifying the distance metric to use `c("haversine", "euclidean")`.
#' @param R A numeric value representing the radius of the Earth in kilometers. The default is 6371 km.
#'
#' @return A list where each element corresponds to a point and contains the indices of its neighbors, excluding the point itself.
#'
#' @details
#' This function uses kd-tree (via `nabor` package) for efficient spatial searches. The kd-tree inherently works with Euclidean distances.
#' If `"haversine"` is selected, the function first converts geographic coordinates to 3D Cartesian coordinates before constructing the kd-tree.
#'
#' @examples
#' set.seed(123)
#' coords <- matrix(runif(20, min = -180, max = 180), ncol = 2)
#'
#' # Compute neighbors using kd-tree
#' neighbors <- compute_neighbors_kdtree(coords, thin_dist = 10,)
#'
#' @export
compute_neighbors_kdtree <- function(coordinates, thin_dist, k = NULL, distance = c("haversine", "euclidean"), R = 6371) {
  # Initialize a list to store neighbor indices
  n <- nrow(coordinates)
  neighbor_indices <- vector("list", n)
  distance <- match.arg(distance)
  if (is.null(k)){
    k <- n
  }

  # In kd-tree we can only use the Euclidean distance
  # When asked for the Haversine distance we transform coordinates
  # Convert geographic coordinates to Cartesian coordinates if lon lat
  if (distance == "haversine") {
    cartesian_points <- t(apply(coordinates, 1, function(row) lon_lat_to_cartesian(row[1], row[2], R)))
  } else if (distance == "euclidean"){
    cartesian_points <- coordinates
  }

  # Build kd-tree and find neighbors within the specified radius
  kd_tree <- nabor::knn(cartesian_points, k = k, radius = thin_dist)

  # Create a list of neighbor indices excluding self-reference
  for (i in seq_len(n)) {
    neighbors <- kd_tree$nn.idx[i, ]
    neighbor_indices[[i]] <- neighbors[neighbors != 0][-1] # Exclude self
  }

  return(neighbor_indices)
}

#' Compute Neighbors Using Local kd-Trees
#'
#' Divides the search area into a grid of local regions and constructs kd-trees
#' for each region to compute neighbors efficiently. Neighbor regions are
#' also considered to ensure a complete search.
#'
#' @param coordinates A matrix of coordinates to thin, with two columns representing longitude and latitude.
#' @param thin_dist A positive numeric value representing the thinning distance in kilometers.
#' @param distance A character string specifying the distance metric to use `c("haversine", "euclidean")`.
#' @param R A numeric value representing the radius of the Earth in kilometers. The default is 6371 km.
#' @param n_cores An integer specifying the number of cores to use for parallel processing. The default is 1.
#'
#' @return A list where each element corresponds to a point and contains the indices of its neighbors, excluding the point itself.
#'
#' @examples
#' set.seed(123)
#' coords <- matrix(runif(20, min = -180, max = 180), ncol = 2)
#'
#' # Compute neighbors using local kd-trees with Euclidean distance
#' neighbors <- compute_neighbors_local_kdtree(coords, thin_dist = 10, n_cores = 1)
#'
#' @importFrom foreach %dopar%
#' @export
compute_neighbors_local_kdtree <- function(coordinates, thin_dist, distance = c("haversine", "euclidean"), R = 6371, n_cores = 1) {
  # Initialize a list to store neighbor indices
  n <- nrow(coordinates)
  neighbor_indices <- vector("list", n)
  distance <- match.arg(distance)

  # In kd-tree we can only use the Euclidean distance
  # When asked for the Haversine distance we transform coordinates
  # Convert geographic coordinates to Cartesian coordinates if lon lat
  if (distance == "haversine") {
    cartesian_points <- t(apply(coordinates, 1, function(row) lon_lat_to_cartesian(row[1], row[2], R)))
    # Assign points to 3D grid using integer indexing
    grid_coords <- floor(cartesian_points / thin_dist)

    # Create a list to store points by grid cell
    #grid_dict <- split(seq_len(n), list(grid_coords[,1], grid_coords[,2], grid_coords[,3]), drop = TRUE)
    cell_keys <- paste(grid_coords[, 1], grid_coords[, 2], grid_coords[, 3], sep = ".")
    grid_dict <- split(seq_len(n), cell_keys)

  } else if (distance == "euclidean"){
    cartesian_points <- coordinates

    # Assign points to 2D grid
    grid_coords <- floor(cartesian_points / thin_dist)
    grid_dict <- split(seq_len(n), list(grid_coords[,1], grid_coords[,2], 0), drop = TRUE)
  }

  #Find neighbors for each point
  process_grid_cell <- function(cell_key) {
    cell_ids <- grid_dict[[cell_key]]

    if (length(cell_ids) == 0){
      return(NULL)
    }

    # Find neighbor cells
    grid_x <- as.numeric(strsplit(cell_key, "\\.")[[1]][1])
    grid_y <- as.numeric(strsplit(cell_key, "\\.")[[1]][2])
    grid_z <- as.numeric(strsplit(cell_key, "\\.")[[1]][3])

    neighbor_keys <- c()
    for (dx in -1:1) {
      for (dy in -1:1) {
        for (dz in -1:1) {
          neighbor_keys <- c(neighbor_keys, paste(grid_x + dx, grid_y + dy, grid_z + dz, sep = "."))
        }
      }
    }

    # Collect all points in query cell and neighbor cells
    neighbor_ids <- unlist(grid_dict[neighbor_keys], use.names = FALSE)

    # Build kd-tree and find neighbors within the specified radius
    cell_points <- cartesian_points[cell_ids, , drop = FALSE]
    combined_points <- cartesian_points[neighbor_ids, , drop = FALSE]

    kd_tree <- nabor::knn(data = combined_points, query = cell_points, k = nrow(combined_points), radius = thin_dist)

    # Loop through each point in the current grid cell
    result <- vector("list", length(cell_ids))
    for (i in seq_along(cell_ids)) {
      neighbors <- kd_tree$nn.idx[i, ]

      # Remove self-reference and map back to original indices
      result[[i]] <- neighbor_ids[neighbors[neighbors != 0]][-1]
    }

    names(result) <- cell_ids
    return(result)
  }

  if (n_cores > 1) { # Parallelization
    if (parallel::detectCores() <= n_cores) {
      stop("`n_cores` must be smaller than the available number of cores. Check with `detectCores()`.")
    }

    cl <- parallel::makeCluster(n_cores)
    #parallel::clusterExport(cl, c("coordinates", "cartesian_points", "thin_dist"), envir = environment())
    doParallel::registerDoParallel(cl)

    results <- foreach::foreach(cell_key = names(grid_dict), .packages = "nabor") %dopar% {
      process_grid_cell(cell_key)
    }

    parallel::stopCluster(cl)

    # Merge results
    for (res in results) {
      if (!is.null(res)) {
        for (j in names(res)) {
          neighbor_indices[[as.integer(j)]] <- res[[j]]
        }
      }
    }
  } else {  # Single-core execution
    for (cell_key in names(grid_dict)) {
      res <- process_grid_cell(cell_key)
      if (!is.null(res)) {
        for (j in names(res)) {
          neighbor_indices[[as.integer(j)]] <- res[[j]]
        }
      }
    }
  }
  return(neighbor_indices)
}

set.seed(123)

coords <- data.frame(
  lon = c(-122.4194, -122.4195, -122.4196, -122.4197),
  lat = c(37.7749, 37.7740, 37.7741, 37.7750)
)
coords <- as.matrix(coords)

coords2 <- matrix(runif(100, 0, 0.2), ncol=2)

test_that("thin_points returns a GeoThinned object", {
  result <- thin_points(data.frame(lon = runif(10), lat = runif(10)), method = "distance", thin_dist = 1)
  expect_s3_class(result, "GeoThinned")
})

test_that("thin_points handles invalid input", {
  expect_error(distance_thinning(NULL), "`coordinates` must be a matrix")
  expect_error(distance_thinning(coords, thin_dist = -2), "`thin_dist` must be a positive number.")
  expect_error(thin_points(data.frame(lon = runif(10), lat = runif(10)),
                           trials = -5), "`trials` must be a positive integer.")
  expect_error(thin_points(NULL, all_trials = "Yes"), "`all_trials` must be a logical value.")

  expect_error(thin_points(data.frame(lon = runif(10), lat = runif(10)),
                           group_col = "nonexistent"), "Specified grouping column does not exist in the data.")
})

test_that("distance_thinning (brute method) works with geographic coordinates", {
  result <- distance_thinning(coords, search_type = "brute", thin_dist = 0.1, trials = 1)

  expect_length(result[[1]], nrow(coords))
  expect_true(sum(result[[1]]) < nrow(coords)) # Should thin some points
})

test_that("distance_thinning (brute method) works computing Euclidean distance", {
  result <- distance_thinning(coords, search_type = "brute", thin_dist = 1, trials = 1, distance = "euclidean")

  expect_length(result[[1]], nrow(coords))
  expect_true(sum(result[[1]]) < nrow(coords)) # Should thin some points
})

test_that("distance_thinning (brute method) returns the exact number of points when target_points is not NULL", {
  set.seed(2024)
  result <- suppressMessages(distance_thinning(coords, search_type = "brute", thin_dist = 0.1, trials = 3, all_trials = FALSE, target_points = 2, distance = "euclidean"))
  result <- suppressMessages(distance_thinning(coords, search_type = "brute", thin_dist = 0.1, trials = 3, all_trials = FALSE, target_points = 2))

  expect_length(result[[1]], 4)
  expect_equal(sum(result[[1]]), 2)
})

test_that("distance_thinning returns all trials correctly", {
  result <- distance_thinning(coords, search_type = "brute", thin_dist = 0.1, trials = 3, all_trials = TRUE)
  expect_equal(length(result), 3)
  expect_true(all(sapply(result, length) == nrow(coords)))
})

test_that("distance_thinning handles small datasets", {
  small_coords <- coords[1:2, ]
  result <- distance_thinning(small_coords, search_type = "brute", thin_dist = 0.1, trials = 1)
  expect_length(result[[1]], 2)
})

test_that("kd-tree thinning works", {
  result <- distance_thinning(coords, search_type = "kd_tree", thin_dist = 0.1, trials = 1)
  expect_length(result[[1]], nrow(coords))
  expect_true(sum(result[[1]]) < nrow(coords)) # Should thin some points

  result <- distance_thinning(coords, search_type = "kd_tree", thin_dist = 0.1, trials = 1, distance = "euclidean")
  expect_length(result[[1]], nrow(coords))
  expect_true(sum(result[[1]]) < nrow(coords)) # Should thin some points
})

test_that("kd-tree k_max estimation", {
  result <- distance_thinning(coords, search_type = "k_estimation", thin_dist = 0.1, trials = 1)
  expect_length(result[[1]], nrow(coords))
  expect_true(sum(result[[1]]) < nrow(coords)) # Should thin some points
})

test_that("local kd-trees and parallelization", {
  result <- distance_thinning(coords, search_type = "local_kd_tree", thin_dist = 0.1, trials = 1, n_cores = 2)
  expect_length(result[[1]], nrow(coords))
  expect_true(sum(result[[1]]) < nrow(coords)) # Should thin some points

  result <- distance_thinning(coords, search_type = "local_kd_tree", thin_dist = 0.1, trials = 1, distance = "euclidean")
  expect_length(result[[1]], nrow(coords))
  expect_true(sum(result[[1]]) < nrow(coords)) # Should thin some points

  expect_error(distance_thinning(coords, search_type = "local_kd_tree", thin_dist = 0.1, trials = 1, n_cores = 100),
               "`n_cores` must be smaller than the available number of cores")

})

test_that("All methods find same neighbors", {
  thin_dist <- 10
  brute <- compute_neighbors_brute(coords2, thin_dist)
  kd_tree <- compute_neighbors_kdtree(coords2, thin_dist)
  k_estimation <- compute_neighbors_kdtree(coords2, thin_dist, estimate_k_max(coords2, thin_dist))
  local_kd_tree <- compute_neighbors_local_kdtree(coords2, thin_dist)

  brute <- lapply(brute, sort)
  kd_tree <- lapply(kd_tree, sort)
  k_estimation <- lapply(k_estimation, sort)
  local_kd_tree <- lapply(local_kd_tree, sort)

  expect_true(identical(brute, kd_tree))
  expect_true(identical(brute, k_estimation))
  expect_true(identical(brute, local_kd_tree))
})


test_that("distance_thinning works with priority", {
  priority <- 1:nrow(coords)
  result_priority <- distance_thinning(coords, search_type = "brute", thin_dist = 0.1, trials = 1, priority = priority)

  # distance_thinning with wrongly formated priority
  expect_error(distance_thinning(coords, search_type = "brute", thin_dist = 0.1, trials = 1, priority = 1), "'priority' must be a numeric vector with same length as number of points.")
  expect_warning(distance_thinning(coords, search_type = "brute", thin_dist = 0.1, trials = 1, priority = c(NA, 2:nrow(coords))), "NA values found in 'priority'.")

  expect_length(result_priority[[1]], nrow(coords))
  expect_true(sum(result_priority[[1]]) < nrow(coords))
})

#' Spatial Thinning of Points
#'
#' This function performs spatial thinning of geographic points to reduce
#' point density while maintaining spatial representation. Points are thinned
#' based on a specified distance, grid, or decimal precision, with support for
#' multiple trials and optional grouping.
#' By default, only the largest subset is returned. To return all trials, set `all_trials = TRUE`.
#'
#' @param data A data frame or tibble containing the input points to thin. Must contain longitude and latitude columns.
#' @param lon_col Character name of the column with longitude coordinates (default: `"lon"`).
#' @param lat_col Character name of the column with latitude coordinates (default: `"lat"`).
#' @param group_col Character name of the column for grouping points (e.g., species name, year). If `NULL`, no grouping is applied.
#' @param method Thinning method to use. One of `"distance", "grid", "precision"`.
#' @param trials Number of thinning iterations to perform (default: `10`). Must be a positive integer.
#' @param all_trials If `TRUE`, returns results of all attempts; if `FALSE`, returns the best attempt with the most points retained (default: `FALSE`).
#' @param seed Optional; an integer seed for reproducibility of results.
#' @param verbose If `TRUE`, prints progress messages (default: `FALSE`).
#' @param ... Additional arguments passed to specific thinning methods. See Details.
#'
#' @return A `GeoThinned` object (S3 class), which contains:
#' \itemize{
#'   \item `retained`: A list of logical vectors (one per trial) indicating retained points.
#'   \item `original_data`: The original input dataset.
#'   \item `method`: The thinning method used.
#'   \item `params`: A list of the thinning parameters used.
#' }
#' By default, `thin_points()` returns only the trial with the largest number of retained points.
#' To access all thinning trials, set `all_trials = TRUE`; otherwise, functions such as `largest()` and `get_trial()` will always refer to the same subset.
#'
#' @details
#' The following thinning methods are available:
#'
#' \describe{
#'  \item{`"distance"`}{Forces a specific minimum distance between points.}
#'  \item{`"grid"`}{Applies a grid-based thinning method.}
#'  \item{`"precision"`}{Utilizes precision-based thinning.}
#' }
#'
#' \strong{Distance-based thinning}
#'
#' The specific parameters for distance-based thinning are:
#'
#' \describe{
#'  \item{`thin_dist`}{A positive numeric value representing the thinning distance in kilometers.}
#'  \item{`search_type`}{A character string indicating the neighbor search method 'c("local_kd_tree", "k_estimation", "kd_tree", "brute")'. The defult value is 'local_kd_tree'.}
#'  \item{`distance`}{Distance metric to use 'c("haversine", "euclidean")'. Default is Haversine for geographic coordinates.}
#'  \item{`R`}{The radius of the Earth in kilometers. Default is 6371 km.}
#'  \item{`target_points`}{Optional integer specifying the number of points to retain. If 'NULL' (default), the function tries to maximize the number of points retained.}
#'  \item{`n_cores`}{Number of cores for parallel processing (only for '"local_kd_tree"'). Default is 1.}
#' }
#'
#' \strong{Grid-based thinning}
#'
#' The specific parameters for grid-based thinning are:
#'
#' \describe{
#'  \item{`thin_dist`}{A positive numeric value representing the thinning distance in kilometers.}
#'  \item{`resolution`}{A numeric value representing the resolution (in degrees) of the raster grid. If provided, this takes priority over 'thin_dist'.}
#'  \item{`origin`}{A numeric vector of length 2 (e.g., 'c(0, 0)'), specifying the origin of the raster grid (optional).}
#'  \item{`raster_obj`}{An optional 'terra::SpatRaster' object to use for grid thinning. If provided, the raster object will be used instead of creating a new one.}
#'  \item{`n`}{A positive integer specifying the maximum number of points to retain per grid cell (default: 1).}
#'  \item{`crs`}{An optional CRS (Coordinate Reference System) to project the coordinates and raster (default WGS84, 'epsg:4326'). This can be an EPSG code, a PROJ.4 string, or a 'terra::crs' object.}
#'  \item{`priority`}{A numeric vector of the same length as the number of points with numerical values indicating the priority of each point. Instead of eliminating points randomly, higher values are preferred during thinning.}
#' }
#'
#' \strong{Precision-based thinning}
#'
#' The specific parameters for precision-based thinning are:
#'
#' \describe{
#'  \item{`precision`}{A positive integer specifying the number of decimal places to which coordinates should be rounded. Default is 4.}
#'  \item{`priority`}{A numeric vector of the same length as the number of points with numerical values indicating the priority of each point. Instead of eliminating points randomly, higher values are preferred during thinning.}
#' }
#'
#' For more information on specific thinning methods and inputs, refer to their respective documentation:
#'
#' \itemize{
#'  \item `distance_thinning()`
#'  \item `grid_thinning()`
#'  \item `precision_thinning()`
#' }
#'
#' @examples
#' # Basic usage
#' set.seed(123)
#' sample_data <- data.frame(
#'   lon = runif(100, -10, 10),
#'   lat = runif(100, -5, 5)
#' )
#'
#' result <- thin_points(sample_data, method = "distance", thin_dist = 100)
#'
#' # Grouped thinning
#' sample_data$species <- sample(c("A", "B"), 100, replace = TRUE)
#' grouped_result <- thin_points(sample_data, group_col = "species",
#'                               method = "distance", thin_dist = 100)
#'
#' @export
thin_points <- function(data, lon_col = "lon", lat_col = "lat", group_col = NULL,
                        method = c("distance", "grid", "precision"),
                        trials = 10, all_trials = FALSE,
                        seed = NULL, verbose = FALSE, ...) {

  # Match the method argument
  method <- match.arg(method)

  # Validate inputs
  if (!is.numeric(trials) || length(trials) != 1 || trials <= 0) {
    stop("`trials` must be a positive integer.")
  }
  if (!is.logical(all_trials) || length(all_trials) != 1) {
    stop("`all_trials` must be a logical value.")
  }

  # Set seed for reproducibility
  if (!is.null(seed)) {
    set.seed(seed)
  }

  # Check input data and columns
  # Identify longitude and latitude columns
  if (is.null(lon_col) || is.null(lat_col)) {
    # Check if columns 1 and 2 are numeric
    if (!is.numeric(data[, 1]) || !is.numeric(data[, 2])) {
      stop("Error: The first two columns must be numeric. Please specify lon_col and lat_col arguments.")
    }
    # Rename the columns
    lon_col <- "lon"
    lat_col <- "lat"
    colnames(data)[1:2] <- c(lon_col, lat_col)
  }
  if (!lon_col %in% colnames(data) || !lat_col %in% colnames(data)) {
    stop("Specified longitude or latitude columns do not exist in the data.")
  }
  if (!is_lonlat(data[,lon_col], data[, lat_col])) {
    warning(
      "Input coordinates lie outside the typical longitude/latitude global ranges. ",
      "GeoThinneR assumes decimal degrees when using the Haversine distance."
    )
  }
  if (!is.null(group_col) && !(group_col %in% colnames(data))) {
    stop("Specified grouping column does not exist in the data.")
  }

  # Start time tracking
  start_time <- Sys.time()
  if (verbose) cat("Starting spatial thinning at", format(Sys.time()), "\n")

  # Convert data to data frame if it's a tibble or matrix
  is_tbl <- inherits(data, "tbl") | inherits(data, "tbl_df")
  is_mat <- inherits(data, "matrix")
  if (is_tbl | is_mat) {
    data <- as.data.frame(data)
  }

  # Initialize results container
  exported_trials <- ifelse(all_trials, trials, 1)

  # Function to perform thinning
  perform_thinning <- function(data_subset) {
    coordinates <- as.matrix(data_subset[, c(lon_col, lat_col), drop = FALSE])
    if (verbose) cat("Thinning using method:", method, "\n")

    # Call the appropriate thinning method
    kept_points <- switch(
      method,
      "distance"  = distance_thinning(coordinates, trials = trials, all_trials = all_trials, ...),
      "grid"      = grid_thinning(coordinates, trials = trials, all_trials = all_trials, ...),
      "precision" = precision_thinning(coordinates, trials = trials, all_trials = all_trials, ...),
      stop("Invalid thinning method specified.")
    )

    if (verbose) cat("Thinning process completed.\n")
    return(kept_points)
  }

  # Thinning process
  # If group_col is provided, split the data by group
  if (!is.null(group_col)) {
    unique_groups <- unique(data[[group_col]])

    # Preallocate a list of logical vectors for each trial
    kept_points <- vector("list", exported_trials)
    for (i in seq_len(exported_trials)) {
      kept_points[[i]] <- rep(FALSE, nrow(data))
    }

    for (group in unique_groups) {
      if (verbose) cat("Processing group:", group, "\n")
      group_indices <- which(data[[group_col]] == group)
      group_data <- data[group_indices, , drop = FALSE]

      group_kept <- perform_thinning(group_data)

      for (i in seq_len(exported_trials)) {
        kept_points[[i]][group_indices] <- group_kept[[i]]
      }
    }
  } else {
    kept_points <- perform_thinning(data)
  }

  # End time tracking and calculate duration
  execution_time <- Sys.time() - start_time
  if (verbose) cat("Total execution time:", round(execution_time, 2), "seconds\n")

  all_params <- list(...)
  all_params$lon_col <- lon_col
  all_params$lat_col <- lat_col
  all_params$group_col <- group_col
  all_params$trials <- trials
  all_params$all_trials <- all_trials
  all_params$seed <- seed

  result <- as_GeoThinned(
    retained = kept_points[seq_len(exported_trials)],
    method = method,
    params = all_params,
    original_data = data
  )

  return(result)
}

---
title: "Getting started with GeoThinneR"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with GeoThinneR}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
run_code <- requireNamespace("sf", quietly = TRUE) &
  requireNamespace("terra", quietly = TRUE) &
  requireNamespace("ggplot2", quietly = TRUE)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = run_code
)
```

# Overview

**GeoThinneR** is an R package designed for efficient spatial thinning of point data, particularly useful for species occurrence data and other spatial analyses. It implements optimized neighbor search algorithms based on *kd*-trees, providing an **efficient, scalable, and user-friendly** approach. When working with large-scale datasets containing hundreds of thousands of points, applying thinning manually is not possible, and existing tools in R providing thinning methods for species distribution modeling often lack scalability and computational efficiency.

Spatial thinning is widely used across many disciplines to **reduce the density of spatial point data** by selectively removing points based on specified criteria, such as threshold spacing or grid cells. This technique is applied in fields like remote sensing, where it improves data visualization and processing, as well as in geostatistics, and spatial ecological modeling. Among these, species distribution modeling is a key area where thinning plays a crucial role in mitigating the effects of spatial sampling bias.

This vignette provides a step-by-step guide to thinning with **GeoThinneR** using both simulated and real-world datasets.

# Setup and loading data

To get started with **GeoThinneR**, we begin by loading the packages used in this vignette:

```{r setup, message=FALSE, warning=FALSE}
library(GeoThinneR)
library(terra)
library(sf)
library(ggplot2)
```

We'll use two dataset to demonstrate the use of **GeoThinneR**:

1. A **simulated a dataset** with `n = 2000` random occurrences randomly distributed within a 20 x 10 degree area for two artificial species.
2. A **real-world dataset** containing loggerhead sea turtle (*Caretta caretta*) occurrences in the Mediterranean Sea downloaded from [GBIF](https://doi.org/10.15468/dl.9jcjrm) and included with the package.

```{r load data, message=FALSE, results='hide'}
# Set seed for reproducibility
set.seed(123)

# Simulated dataset
n <- 2000
sim_data <- data.frame(
  lon = runif(n, min = -10, max = 10),
  lat = runif(n, min = -5, max = 5),
  sp = sample(c("sp1", "sp2"), n, replace = TRUE)
)

# Load real-world occurrence dataset
data("caretta")

# Load Mediterranean Sea polygon
medit <- system.file("extdata", "mediterranean_sea.gpkg", package = "GeoThinneR")
medit <- sf::st_read(medit)
```

Let's visualize the datasets before thinning:

```{r plot loaded data, fig.show="hold", out.width="50%", echo = FALSE}
ggplot() +
  geom_point(data = sim_data, aes(x = lon, y = lat, shape = sp), color = "#5183B3") +
  scale_shape_manual(values = c(sp1 = 16, sp2 = 17)) +
  xlab("Longitude") + ylab("Latitude") +
  ggtitle("Simulated Species Occurrences") +
  theme_minimal()

ggplot() +
  geom_sf(data = medit, color = "#353839", fill = "antiquewhite", alpha = 0.7)  +
  geom_point(data = caretta, aes(x = decimalLongitude, y = decimalLatitude), color = "#5183B3", alpha = 1) +
  xlab("Longitude") + ylab("Latitude") +
  ggtitle("C. caretta Mediterranean Sea Occurrences") +
  theme(
      panel.grid.major = element_line(color = gray(.5),linetype = "dashed",linewidth = 0.5),
      panel.background = element_rect(fill = "white"),
      axis.title = element_blank(),
      legend.position = "bottom"
    )
```

# Quick start guide

Now that our data is ready, let's use **GeoThinneR** for spatial thinning. All methods are accessed through the `thin_points()` function, which requires, at a minimum, a dataset containing georeferenced points (WGS84 CRS - EPSG:4326) in `matrix`, `data.frame`, or `tibble` format. The query dataset is provided through the `data` argument, and with `lon_col` and `lat_col` parameters, we can define the column names containing longitude and latitude (or x, y) coordinates. Default values are "lon" and "lat", and if not present, it will take the first two columns by default.

The **thinning method** is defined using the `method` parameter, with the following options:

- `"distance"`: Distance-based thinning  
- `"grid"`: Grid-based thinning  
- `"precision"`: Precision-based thinning  

The **thinning constraint** (distance, grid resolution or decimal precision) depends on the method chosen (these will be detailed in the next section). Regardless of the method, the algorithm selects which points to retain and remove. Since this selection is non-deterministic, you can run **multiple randomized thinning trials** to maximize the number of retained points.

Let’s quickly thin the simulated dataset using a simple distance-based method:

```{r}
# Apply spatial thinning to the simulated data
quick_thin <- thin_points(
  data = sim_data,     # Dataframe with coordinates
  lon_col = "lon",     # Longitude column name
  lat_col = "lat",     # Latitude column name
  method = "distance", # Method for thinning
  thin_dist = 20,      # Thinning distance in km,
  trials = 5,          # Number of trials
  all_trials = TRUE,   # Return all trials
  seed = 123           # Seed for reproducibility
)

quick_thin
```

This command applies spatial thinning ensuring no two retained points are closer than 20 km (`thin_dist`). We specified to run 5 random trials (`trials`). By default, it returns only the trial that retains the largest subset of points, but because we set `all_trials = TRUE`, all trials are returned.

### Exploring the GeoThinned object

The output of `thin_points()` is a `GeoThinned` object, which contains:
- `x$retained`: a list of logical vectors indicating which points were retained in each trial
- `x$method`: the method used for thinning
- `x$params`: the parameters used for thinning
- `x$original_data`: the original dataset

You can apply basic methods for inspecting the object such as `print()`, `summary()`, and `plot()`. The `summary()` function provides detailed information about the thinning process, including the number of points retained after thinning, the nearest neighbor distance (NND), and the spatial coverage of the retained points. By default, the summary shows the metrics for the trial that retains the largest number of points, but you can also specify a specific trial using the `trial` argument.

```{r}
summary(quick_thin)
```

We see that from 2000, we retained the 69.5% of points (1390), and the minimum distance between points has increased without reducing too much the spatial coverage or extent of the points compared to the original dataset.

```{r}
plot(quick_thin)
```


If we check the number of points retained in each trial, we can see that the 5 trials return different numbers of points retained. This is due to the randomness of the algorithm when selecting which points to retain. You can use `seed` for reproducibility. By setting, `all_trials = FALSE` it will just return the trial that kept the most points.

```{r}
# Number of kept points in each trial
sapply(quick_thin$retained, sum)
```

You can access the thinned dataset that retained the maximum number of points using the `largest()` function.

```{r}
head(largest(quick_thin))
```

If you want to access points from a specific trial, you can use the `get_trial()` function, which returns the thinned dataset for that trial.

```{r}
head(get_trial(quick_thin, trial = 2)) 
```

Just in case you want to get the points that were removed from the original dataset, you can use the `retained` list to index the original data. The `retained` list contains logical vectors indicating which points were retained in each trial. You can use this to filter the original data and get the removed points.

```{r}
trial <- 2
head(quick_thin$original_data[!quick_thin$retained[[trial]], ])
```

Also, you can use `as_sf()` to convert the thinned dataset to a `sf` object, which is useful for plotting and spatial analysis.

```{r}
sf_thinned <- as_sf(quick_thin)
class(sf_thinned)
```

### Real-world example

GeoThinneR includes a dataset of loggerhead sea turtle (*Caretta caretta*) occurrence records in the Mediterranean Sea, sourced from GBIF and filtered. We now apply a 30 km thinning distance, returning only the largest trial (the one with the maximum number of points retained, `all_trials` set to `FALSE`).

```{r}
# Apply spatial thinning to the real data
caretta_thin <- thin_points(
  data = caretta, # We will not specify lon_col, lat_col as they are in position 1 and 2
  lat_col = "decimalLatitude",
  lon_col = "decimalLongitude",
  method = "distance",
  thin_dist = 30, # Thinning distance in km,
  trials = 5,
  all_trials = FALSE,
  seed = 123
)

# Thinned dataframe stored in the first element of the retained list
identical(largest(caretta_thin), get_trial(caretta_thin, 1))
```
```{r, echo = FALSE}
ggplot() +
  geom_sf(data = medit, color = "#353839", fill = "antiquewhite", alpha = 0.7)  +
  geom_point(data = caretta, aes(x = decimalLongitude, y = decimalLatitude),color = "#EB714B", alpha = 0.5) +
  geom_point(data = largest(caretta_thin), aes(x = decimalLongitude, y = decimalLatitude),color = "#5183B3", alpha = 1) +
  xlab("Longitude") + ylab("Latitude") +
  ggtitle("C. caretta Mediterranean Sea Occurrences") +
  theme(
      panel.grid.major = element_line(color = gray(.5),linetype = "dashed",linewidth = 0.5),
      panel.background = element_rect(fill = "white"),
      axis.title = element_blank(),
      legend.position = "bottom"
    )
```

As you can see, with **GeoThinneR**, it's very easy to apply spatial thinning to your dataset. In the next section we explain the different thinning methods and how to use them, and also the implemented optimizations for large-scale datasets.

# Thinning methods

**GeoThinneR** offers different algorithms for spatial thinning. In this section, we will show how to use each thinning method, highlighting its unique parameters and features.

## Distance-based thinning

Distance-based thinning (`method="distance"`) ensures that all retained points are separated by at least a user-defined minimum distance `thin_dist` (in km). For geographic coordinates, using `distance = "haversine"` (default) computes great-circle distances, or alternatively, the Euclidean distance between 3D Cartesian coordinates (depending on the method) to identify neighboring points within the distance threshold. The algorithm iterates through the dataset and removes points that are too close to each other.

In general, the choice of search method depends on the size of the dataset and the desired thinning distance:

- `search_type="brute"`: Useful for small datasets, low scalability.
- `search_type="kd_tree"`: More memory-efficient, better scalability for medium-sized datasets.
- `search_type="local_kd_tree"` (default): Faster than `"kd_tree"` and reduces memory usage. Can be parallelized to improve speed.
- `search_type="k_estimation"`: Very fast when the number of points to remove is low or not very clustered, if not similar or worse than `"local_kd_tree"`.

You can find the benchmarking results for each method in the package manuscript (see references).

```{r}
dist_thin <- thin_points(sim_data, method = "distance", thin_dist = 25, trials = 3)
plot(dist_thin)
```


### Brute force

This is the most common greedy method for calculating the distance between points (`search_type="brute"`), as it directly computes all pairwise distances and retains points that are far enough apart based on the thinning distance. By default, it computes the Haversine distance using the `rdist.earth()` function from the **fields** package. If `distance = "euclidean"`, it computes the Euclidean distance instead. The main advantage of this method is that it computes the full distance matrix between your points. However, the main disadvantage is that it is very time and memory consuming, as it requires computing all pairwise comparisons.

```{r}
brute_thin <- thin_points(
  data = sim_data,
  method = "distance",
  search_type = "brute",
  thin_dist = 20,
  trials = 10,
  all_trials = FALSE,
  seed = 123
)
nrow(largest(brute_thin))
```

### Traditional *kd*-tree

A *kd*-tree is a binary tree that partitions space into regions to optimize nearest neighbor searches. The `search_type="kd_tree"` method uses the **nabor** R package to implement *kd*-trees via the **libnabo** library. This method is more memory-efficient than brute force. Since this implementation of *kd*-trees is based on Euclidean distance, when working with geographic coordinates (`distance = "haversine"`), **GeoThinneR** transforms the coordinates into XYZ Cartesian coordinates. Although this method scales better than the brute-force method for large datasets, since we want to identify all neighboring points of each query point, its performance can still be improved for very large datasets.

```{r}
kd_tree_thin <- thin_points(
  data = sim_data,
  method = "distance",
  search_type = "kd_tree",
  thin_dist = 20,
  trials = 10,
  all_trials = FALSE,
  seed = 123
)
nrow(largest(kd_tree_thin))
```

### Local *kd*-trees

To reduce the complexity of searching within a single huge *kd*-tree containing all points from the dataset, we propose a simple modification: creating small local *kd*-trees by dividing the spatial domain into multiple smaller regions, each with its own tree (`search_type="local_kd_tree"`). This notably reduces memory usage, allowing this thinning method to be used for very large datasets on personal computers. Additionally, you can run in parallel this method to improve speed using the `n_cores` parameter.

```{r}
local_kd_tree_thin <- thin_points(
  data = sim_data,
  method = "distance",
  search_type = "local_kd_tree",
  thin_dist = 20,
  trials = 10,
  all_trials = FALSE,
  seed = 123
)
nrow(largest(local_kd_tree_thin))
```

### Restricted neighbor search

Another way to reduce the computational complexity of *kd*-tree neighbor searches is by limiting the number of neighbors to find. In the `nabor::knn()` function, we can specify the number of neighbors to search for (`k`). For spatial thinning, we need to identify all neighbors for each point, which increases computational cost by setting `k` to the total number of points (`n`). However, with the `search_type="k_estimation"` method, we can estimate the maximum number of neighbors a point can have based on the spatial density of points, thereby reducing the `k` parameter. This method is particularly useful when datasets are not highly clustered, as it estimates a smaller `k` value, significantly reducing computational complexity.

```{r}
k_estimation_thin <- thin_points(
  data = sim_data,
  method = "distance",
  search_type = "k_estimation",
  thin_dist = 20,
  trials = 10,
  all_trials = FALSE,
  seed = 123
)
nrow(largest(k_estimation_thin))
```

## Grid-based thinning

Grid sampling is a standard method where the area is divided into a grid, and `n` points are sampled from each grid cell. This method is very fast and memory-efficient. The only drawback is that you cannot ensure a minimum distance between points. There are two main ways to apply grid sampling in **GeoThinneR**: (i) define the characteristics of the grid, or (ii) pass your own grid as a raster (`terra::SpatRaster`).

For the first method, you can use the `thin_dist` parameter to define the grid cell size (the distance in km will be approximated to degrees to define the grid cell size, one degree roughly 111.32 km), or you can pass the resolution of the grid (e.g., `resolution = 0.25` for 0.25x0.25-degree cells). If you want to align the grid with external data or covariate layers, you can pass the `origin` argument as a tuple of two values (e.g., `c(0, 0)`). Similarly, you can specify the coordinate reference system (CRS) of your grid (`crs`).

```{r}
system.time(
grid_thin <- thin_points(
  data = sim_data,
  method = "grid",
  resolution = 1,
  origin = c(0, 0),
  crs = "epsg:4326",
  n = 1, # one point per grid cell
  trials = 10,
  all_trials = FALSE,
  seed = 123
))
nrow(largest(grid_thin))
```

Alternatively, you can pass a `terra::SpatRaster` object, and that grid will be used for the thinning process.

```{r, message=FALSE, warning=FALSE}
rast_obj <- terra::rast(xmin = -10, xmax = 10, ymin = -5, ymax = 5, res = 1)
system.time(
grid_raster_thin <- thin_points(
  data = sim_data,
  method = "grid", 
  raster_obj = rast_obj,
  trials = 10,
  n = 1,
  all_trials = FALSE,
  seed = 123
))
nrow(largest(grid_raster_thin))
```
```{r, echo = FALSE}
rast_obj <- terra::as.polygons(rast_obj)
rast_obj <- sf::st_as_sf(rast_obj)
ggplot() +
  geom_sf(data = rast_obj, color = "#353839") +
  geom_point(data = sim_data, aes(x = lon, y = lat),color = "#EB714B", alpha = 0.5) +
  geom_point(data = largest(grid_raster_thin), aes(x = lon, y = lat),color = "#5183B3", alpha = 1) +
  labs(title = "Grid-based thinning results",
       x = "Longitude",
       y = "Latitude",
       caption = "Blue points are kept, red points are deleted") +
  theme_minimal()
```

## Precision thinning

In this approach, coordinates are rounded to a certain precision to remove points that fall too close together. After removing points based on coordinate precision, the coordinate values are restored to their original locations. This is the simplest method and is more of a filtering when working with data from different sources with varying coordinate precision. To use it, you need to define the `precision` parameter, indicating the number of decimals to which the coordinates should be rounded.

```{r}
system.time(
precision_thin <- thin_points(
  data = sim_data,
  method = "precision", 
  precision = 0,
  trials = 50,
  all_trials = FALSE,
  seed = 123
))
nrow(largest(precision_thin))
```
```{r, echo = FALSE}
ggplot() +
  geom_point(data = sim_data, aes(x = lon, y = lat),  color = "#EB714B", size = 3, alpha = 0.5) +
  geom_point(data = largest(precision_thin), aes(x = lon, y = lat), color = "#5183B3", size = 3) +
  labs(title = "Precision-based thinning results",
       x = "Longitude",
       y = "Latitude",
       caption = "Blue points are kept, red points are deleted") +
  theme_minimal()
```

These are the methods implemented in **GeoThinneR**. Depending on your specific dataset and research needs, one method may be more suitable than others.

# Additional customization features

## Spatial thinning by group

In some cases, your dataset may include different groups, such as species, time periods, areas, or conditions, that you want to thin independently. The `group_col` parameter allows you to specify the column containing the grouping factor, and the thinning will be performed separately for each group. For example, in the simulated data where we have two species, we can use this parameter to thin each species independently:

```{r}
all_thin <- thin_points(
  data = sim_data,
  thin_dist = 100,
  seed = 123
)
grouped_thin <- thin_points(
  data = sim_data,
  group_col = "sp",
  thin_dist = 100,
  seed = 123
)

nrow(largest(all_thin))
nrow(largest(grouped_thin))
```
```{r, echo =FALSE, fig.show="hold",  out.width="50%"}
removed <- sim_data[!all_thin$retained[[1]], ]
removed_group <- sim_data[!grouped_thin$retained[[1]], ]

ggplot() +
  geom_point(data = removed, aes(x = lon, y = lat, shape = sp), color = "#EB714B",  alpha = 0.2) +
  geom_point(data = largest(all_thin), aes(x = lon, y = lat, shape = sp), color = "#5183B3", alpha = 1) +
  scale_shape_manual(values = c(sp1 = 15, sp2 = 17)) +
  ggtitle("Thinning without grouping") +
  theme_minimal()

ggplot() +
  geom_point(data = removed_group, aes(x = lon, y = lat, shape = sp), color = "#EB714B",  alpha = 0.2) +
  geom_point(data = largest(grouped_thin), aes(x = lon, y = lat, shape = sp), color = "#5183B3", alpha = 1) +
  scale_shape_manual(values = c(sp1 = 15, sp2 = 17)) +
  ggtitle("Thinning by group") +
  theme_minimal()
```



## Fixed number of points

What if you need to retain a fixed number of points that best covers the area where your data points are located, or for class balancing? The `target_points` parameter allows you to specify the number of points to keep, and the function will return that number of points spaced as separated as possible. Additionally, you can also set a `thin_dist` parameter so that no points closer than this distance will be retained. Currently, this approach is only implemented using the brute force method (`method = "distance"` and `search_type = "brute"`), so be cautious when applying it to very large datasets.

```{r}
targeted_thin <- thin_points(
  data = caretta,
  lon_col = "decimalLongitude",
  lat_col = "decimalLatitude",
  target_points = 150,
  thin_dist = 30,
  all_trials = FALSE,
  seed = 123,
  verbose = TRUE
)
nrow(largest(targeted_thin))
```
```{r, echo = FALSE}
ggplot() +
  geom_sf(data = medit, color = "#353839", fill = "antiquewhite", alpha = 0.7)  +
  geom_point(data = largest(targeted_thin), aes(x = decimalLongitude, y = decimalLatitude),color = "#5183B3", alpha = 1) +
  xlab("Longitude") + ylab("Latitude") +
  ggtitle("Exact number of occurrences") +
  theme(
      panel.grid.major = element_line(color = gray(.5),linetype = "dashed",linewidth = 0.5),
      panel.background = element_rect(fill = "white"),
      axis.title = element_blank(),
      legend.position = "bottom"
    )
```

## Select points by priority

In some scenarios, you may want to prioritize certain points based on a specific variable, such as coordinate uncertainty, data quality, or any other variable. The `priority` parameter allows you to assign a numeric weight to each point, where higher values indicate higher importance, and these points are more likely to be retained during thinning. This feature is available for the `"distance"`, `"grid"` and `"precision"` methods.

For example, in the sea turtle data downloaded from GBIF, there is a column named `coordinateUncertaintyInMeters`. To prioritize records with lower uncertainty, you can invert the values so that smaller uncertainties receive higher priority. A simple way to do this is:

```{r}
# Invert uncertainty to give more priority to lower uncertainty records
priority <- -caretta$coordinateUncertaintyInMeters

# For example, handle NAs by assigning lowest possible priority
priority[is.na(priority)] <- min(priority, na.rm = TRUE) - 1
```

Then apply thinning using the priority weights:

```{r}
# Standard grid thinning (no priority)
grid_thin <- thin_points(
  data = caretta,
  lon_col = "decimalLongitude",
  lat_col = "decimalLatitude",
  method = "grid",
  resolution = 2,
  seed = 123
)

# Grid thinning with priority
priority_thin <- thin_points(
  data = caretta,
  lon_col = "decimalLongitude",
  lat_col = "decimalLatitude",
  method = "grid",
  resolution = 2,
  priority = priority,
  seed = 123
)

mean(largest(grid_thin)$coordinateUncertaintyInMeters, na.rm = TRUE)
mean(largest(priority_thin)$coordinateUncertaintyInMeters, na.rm = TRUE)
```

**Notes**

* Higher `priority` values mean the point is more likely to be retained.
* In `"grid"` and `"precision"` methods, when multiple points fall into the same grid cell or rounded coordinate, the one with highest priority is kept.
* In the `"distance"` method, priority is used to resolve tie situations where multiple points have the same spatial influence; the point with lowest priority is dropped.
* If multiple points have the same priority, one is selected at random.
* By default, if provided `NA` values in `priority`, they are treated as the lowest possible priority.

# Other Packages

The **GeoThinneR** package was inspired by the work of many others who have developed methods and packages for working with spatial data and thinning techniques. Our goal with **GeoThinneR** is to offer additional flexibility in method selection and to address specific needs we encountered while using other packages. We would like to mention other tools that may be suitable for your work:

- **spThin**: The `thin()` function provides a brute-force spatial thinning of data, maximizing the number of retained points through random iterative repetitions, using the Haversine distance.
- **enmSdmX**: Includes the `geoThin()` function, which calculates all pairwise distances between points for thinning purposes.
- **dismo**: The `gridSample()` function samples points using a grid as stratification, providing an efficient method for spatial thinning.

# References

- Aiello‐Lammens, M. E., Boria, R. A., Radosavljevic, A., Vilela, B., and Anderson, R. P. (2015). *spThin: an R package for spatial thinning of species occurrence records for use in ecological niche models*. Ecography, 38(5), 541-545. https://doi.org/10.1111/ecog.01132

- Elseberg, J., Magnenat, S., Siegwart, R., & Nüchter, A. (2012). *Comparison of nearest-neighbor-search strategies and implementations for efficient shape registration*. Journal of Software Engineering for Robotics, 3(1), 2-12. https://hdl.handle.net/10446/86202

- Hijmans, R.J., Phillips, S., Leathwick, J., & Elith, J. (2023). *dismo: Species Distribution Modeling*. R Package Version 1.3-14. https://cran.r-project.org/package=dismo

- Mestre-Tomás J (2025). GeoThinneR: Efficient Spatial Thinning of Species Occurrences. R package version 2.0.0, https://github.com/jmestret/GeoThinneR

- Smith, A. B., Murphy, S. J., Henderson, D., & Erickson, K. D. (2023). *Including imprecisely georeferenced specimens improves accuracy of species distribution models and estimates of niche breadth*. Global Ecology and Biogeography, 32(3), 342-355. https://doi.org/10.1111/geb.13628

# Session info

```{r}
sessionInfo()
```

---
title: "Getting started with GeoThinneR"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with GeoThinneR}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
run_code <- requireNamespace("sf", quietly = TRUE) &
  requireNamespace("terra", quietly = TRUE) &
  requireNamespace("ggplot2", quietly = TRUE)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = run_code
)
```

# Overview

**GeoThinneR** is an R package designed for efficient spatial thinning of point data, particularly useful for species occurrence data and other spatial analyses. It implements optimized neighbor search algorithms based on *kd*-trees, providing an **efficient, scalable, and user-friendly** approach. When working with large-scale datasets containing hundreds of thousands of points, applying thinning manually is not possible, and existing tools in R providing thinning methods for species distribution modeling often lack scalability and computational efficiency.

Spatial thinning is widely used across many disciplines to **reduce the density of spatial point data** by selectively removing points based on specified criteria, such as threshold spacing or grid cells. This technique is applied in fields like remote sensing, where it improves data visualization and processing, as well as in geostatistics, and spatial ecological modeling. Among these, species distribution modeling is a key area where thinning plays a crucial role in mitigating the effects of spatial sampling bias.

This vignette provides a step-by-step guide to thinning with **GeoThinneR** using both simulated and real-world datasets.

# Setup and loading data

To get started with **GeoThinneR**, we begin by loading the packages used in this vignette:

```{r setup, message=FALSE, warning=FALSE}
library(GeoThinneR)
library(terra)
library(sf)
library(ggplot2)
```

We'll use two dataset to demonstrate the use of **GeoThinneR**:

1. A **simulated a dataset** with `n = 2000` random occurrences randomly distributed within a 20 x 10 degree area for two artificial species.
2. A **real-world dataset** containing loggerhead sea turtle (*Caretta caretta*) occurrences in the Mediterranean Sea downloaded from [GBIF](https://doi.org/10.15468/dl.9jcjrm) and included with the package.

```{r load data, message=FALSE, results='hide'}
# Set seed for reproducibility
set.seed(123)

# Simulated dataset
n <- 2000
sim_data <- data.frame(
  lon = runif(n, min = -10, max = 10),
  lat = runif(n, min = -5, max = 5),
  sp = sample(c("sp1", "sp2"), n, replace = TRUE)
)

# Load real-world occurrence dataset
data("caretta")

# Load Mediterranean Sea polygon
medit <- system.file("extdata", "mediterranean_sea.gpkg", package = "GeoThinneR")
medit <- sf::st_read(medit)
```

Let's visualize the datasets before thinning:

```{r plot loaded data, fig.show="hold", out.width="50%", echo = FALSE}
ggplot() +
  geom_point(data = sim_data, aes(x = lon, y = lat, shape = sp), color = "#5183B3") +
  scale_shape_manual(values = c(sp1 = 16, sp2 = 17)) +
  xlab("Longitude") + ylab("Latitude") +
  ggtitle("Simulated Species Occurrences") +
  theme_minimal()

ggplot() +
  geom_sf(data = medit, color = "#353839", fill = "antiquewhite", alpha = 0.7)  +
  geom_point(data = caretta, aes(x = decimalLongitude, y = decimalLatitude), color = "#5183B3", alpha = 1) +
  xlab("Longitude") + ylab("Latitude") +
  ggtitle("C. caretta Mediterranean Sea Occurrences") +
  theme(
      panel.grid.major = element_line(color = gray(.5),linetype = "dashed",linewidth = 0.5),
      panel.background = element_rect(fill = "white"),
      axis.title = element_blank(),
      legend.position = "bottom"
    )
```

# Quick start guide

Now that our data is ready, let's use **GeoThinneR** for spatial thinning. All methods are accessed through the `thin_points()` function, which requires, at a minimum, a dataset containing georeferenced points (WGS84 CRS - EPSG:4326) in `matrix`, `data.frame`, or `tibble` format. The query dataset is provided through the `data` argument, and with `lon_col` and `lat_col` parameters, we can define the column names containing longitude and latitude (or x, y) coordinates. Default values are "lon" and "lat", and if not present, it will take the first two columns by default.

The **thinning method** is defined using the `method` parameter, with the following options:

- `"distance"`: Distance-based thinning  
- `"grid"`: Grid-based thinning  
- `"precision"`: Precision-based thinning  

The **thinning constraint** (distance, grid resolution or decimal precision) depends on the method chosen (these will be detailed in the next section). Regardless of the method, the algorithm selects which points to retain and remove. Since this selection is non-deterministic, you can run **multiple randomized thinning trials** to maximize the number of retained points.

Let’s quickly thin the simulated dataset using a simple distance-based method:

```{r}
# Apply spatial thinning to the simulated data
quick_thin <- thin_points(
  data = sim_data,     # Dataframe with coordinates
  lon_col = "lon",     # Longitude column name
  lat_col = "lat",     # Latitude column name
  method = "distance", # Method for thinning
  thin_dist = 20,      # Thinning distance in km,
  trials = 5,          # Number of trials
  all_trials = TRUE,   # Return all trials
  seed = 123           # Seed for reproducibility
)

quick_thin
```

This command applies spatial thinning ensuring no two retained points are closer than 20 km (`thin_dist`). We specified to run 5 random trials (`trials`). By default, it returns only the trial that retains the largest subset of points, but because we set `all_trials = TRUE`, all trials are returned.

### Exploring the GeoThinned object

The output of `thin_points()` is a `GeoThinned` object, which contains:
- `x$retained`: a list of logical vectors indicating which points were retained in each trial
- `x$method`: the method used for thinning
- `x$params`: the parameters used for thinning
- `x$original_data`: the original dataset

You can apply basic methods for inspecting the object such as `print()`, `summary()`, and `plot()`. The `summary()` function provides detailed information about the thinning process, including the number of points retained after thinning, the nearest neighbor distance (NND), and the spatial coverage of the retained points. By default, the summary shows the metrics for the trial that retains the largest number of points, but you can also specify a specific trial using the `trial` argument.

```{r}
summary(quick_thin)
```

We see that from 2000, we retained the 69.5% of points (1390), and the minimum distance between points has increased without reducing too much the spatial coverage or extent of the points compared to the original dataset.

```{r}
plot(quick_thin)
```


If we check the number of points retained in each trial, we can see that the 5 trials return different numbers of points retained. This is due to the randomness of the algorithm when selecting which points to retain. You can use `seed` for reproducibility. By setting, `all_trials = FALSE` it will just return the trial that kept the most points.

```{r}
# Number of kept points in each trial
sapply(quick_thin$retained, sum)
```

You can access the thinned dataset that retained the maximum number of points using the `largest()` function.

```{r}
head(largest(quick_thin))
```

If you want to access points from a specific trial, you can use the `get_trial()` function, which returns the thinned dataset for that trial.

```{r}
head(get_trial(quick_thin, trial = 2)) 
```

Just in case you want to get the points that were removed from the original dataset, you can use the `retained` list to index the original data. The `retained` list contains logical vectors indicating which points were retained in each trial. You can use this to filter the original data and get the removed points.

```{r}
trial <- 2
head(quick_thin$original_data[!quick_thin$retained[[trial]], ])
```

Also, you can use `as_sf()` to convert the thinned dataset to a `sf` object, which is useful for plotting and spatial analysis.

```{r}
sf_thinned <- as_sf(quick_thin)
class(sf_thinned)
```

### Real-world example

GeoThinneR includes a dataset of loggerhead sea turtle (*Caretta caretta*) occurrence records in the Mediterranean Sea, sourced from GBIF and filtered. We now apply a 30 km thinning distance, returning only the largest trial (the one with the maximum number of points retained, `all_trials` set to `FALSE`).

```{r}
# Apply spatial thinning to the real data
caretta_thin <- thin_points(
  data = caretta, # We will not specify lon_col, lat_col as they are in position 1 and 2
  lat_col = "decimalLatitude",
  lon_col = "decimalLongitude",
  method = "distance",
  thin_dist = 30, # Thinning distance in km,
  trials = 5,
  all_trials = FALSE,
  seed = 123
)

# Thinned dataframe stored in the first element of the retained list
identical(largest(caretta_thin), get_trial(caretta_thin, 1))
```
```{r, echo = FALSE}
ggplot() +
  geom_sf(data = medit, color = "#353839", fill = "antiquewhite", alpha = 0.7)  +
  geom_point(data = caretta, aes(x = decimalLongitude, y = decimalLatitude),color = "#EB714B", alpha = 0.5) +
  geom_point(data = largest(caretta_thin), aes(x = decimalLongitude, y = decimalLatitude),color = "#5183B3", alpha = 1) +
  xlab("Longitude") + ylab("Latitude") +
  ggtitle("C. caretta Mediterranean Sea Occurrences") +
  theme(
      panel.grid.major = element_line(color = gray(.5),linetype = "dashed",linewidth = 0.5),
      panel.background = element_rect(fill = "white"),
      axis.title = element_blank(),
      legend.position = "bottom"
    )
```

As you can see, with **GeoThinneR**, it's very easy to apply spatial thinning to your dataset. In the next section we explain the different thinning methods and how to use them, and also the implemented optimizations for large-scale datasets.

# Thinning methods

**GeoThinneR** offers different algorithms for spatial thinning. In this section, we will show how to use each thinning method, highlighting its unique parameters and features.

## Distance-based thinning

Distance-based thinning (`method="distance"`) ensures that all retained points are separated by at least a user-defined minimum distance `thin_dist` (in km). For geographic coordinates, using `distance = "haversine"` (default) computes great-circle distances, or alternatively, the Euclidean distance between 3D Cartesian coordinates (depending on the method) to identify neighboring points within the distance threshold. The algorithm iterates through the dataset and removes points that are too close to each other.

In general, the choice of search method depends on the size of the dataset and the desired thinning distance:

- `search_type="brute"`: Useful for small datasets, low scalability.
- `search_type="kd_tree"`: More memory-efficient, better scalability for medium-sized datasets.
- `search_type="local_kd_tree"` (default): Faster than `"kd_tree"` and reduces memory usage. Can be parallelized to improve speed.
- `search_type="k_estimation"`: Very fast when the number of points to remove is low or not very clustered, if not similar or worse than `"local_kd_tree"`.

You can find the benchmarking results for each method in the published article (see references).

```{r}
dist_thin <- thin_points(sim_data, method = "distance", thin_dist = 25, trials = 3)
plot(dist_thin)
```


### Brute force

This is the most common greedy method for calculating the distance between points (`search_type="brute"`), as it directly computes all pairwise distances and retains points that are far enough apart based on the thinning distance. By default, it computes the Haversine distance using the `rdist.earth()` function from the **fields** package. If `distance = "euclidean"`, it computes the Euclidean distance instead. The main advantage of this method is that it computes the full distance matrix between your points. However, the main disadvantage is that it is very time and memory consuming, as it requires computing all pairwise comparisons.

```{r}
brute_thin <- thin_points(
  data = sim_data,
  method = "distance",
  search_type = "brute",
  thin_dist = 20,
  trials = 10,
  all_trials = FALSE,
  seed = 123
)
nrow(largest(brute_thin))
```

### Traditional *kd*-tree

A *kd*-tree is a binary tree that partitions space into regions to optimize nearest neighbor searches. The `search_type="kd_tree"` method uses the **nabor** R package to implement *kd*-trees via the **libnabo** library. This method is more memory-efficient than brute force. Since this implementation of *kd*-trees is based on Euclidean distance, when working with geographic coordinates (`distance = "haversine"`), **GeoThinneR** transforms the coordinates into XYZ Cartesian coordinates. Although this method scales better than the brute-force method for large datasets, since we want to identify all neighboring points of each query point, its performance can still be improved for very large datasets.

```{r}
kd_tree_thin <- thin_points(
  data = sim_data,
  method = "distance",
  search_type = "kd_tree",
  thin_dist = 20,
  trials = 10,
  all_trials = FALSE,
  seed = 123
)
nrow(largest(kd_tree_thin))
```

### Local *kd*-trees

To reduce the complexity of searching within a single huge *kd*-tree containing all points from the dataset, we propose a simple modification: creating small local *kd*-trees by dividing the spatial domain into multiple smaller regions, each with its own tree (`search_type="local_kd_tree"`). This notably reduces memory usage, allowing this thinning method to be used for very large datasets on personal computers. Additionally, you can run in parallel this method to improve speed using the `n_cores` parameter.

```{r}
local_kd_tree_thin <- thin_points(
  data = sim_data,
  method = "distance",
  search_type = "local_kd_tree",
  thin_dist = 20,
  trials = 10,
  all_trials = FALSE,
  seed = 123
)
nrow(largest(local_kd_tree_thin))
```

### Restricted neighbor search

Another way to reduce the computational complexity of *kd*-tree neighbor searches is by limiting the number of neighbors to find. In the `nabor::knn()` function, we can specify the number of neighbors to search for (`k`). For spatial thinning, we need to identify all neighbors for each point, which increases computational cost by setting `k` to the total number of points (`n`). However, with the `search_type="k_estimation"` method, we can estimate the maximum number of neighbors a point can have based on the spatial density of points, thereby reducing the `k` parameter. This method is particularly useful when datasets are not highly clustered, as it estimates a smaller `k` value, significantly reducing computational complexity.

```{r}
k_estimation_thin <- thin_points(
  data = sim_data,
  method = "distance",
  search_type = "k_estimation",
  thin_dist = 20,
  trials = 10,
  all_trials = FALSE,
  seed = 123
)
nrow(largest(k_estimation_thin))
```

## Grid-based thinning

Grid sampling is a standard method where the area is divided into a grid, and `n` points are sampled from each grid cell. This method is very fast and memory-efficient. The only drawback is that you cannot ensure a minimum distance between points. There are two main ways to apply grid sampling in **GeoThinneR**: (i) define the characteristics of the grid, or (ii) pass your own grid as a raster (`terra::SpatRaster`).

For the first method, you can use the `thin_dist` parameter to define the grid cell size (the distance in km will be approximated to degrees to define the grid cell size, one degree roughly 111.32 km), or you can pass the resolution of the grid (e.g., `resolution = 0.25` for 0.25x0.25-degree cells). If you want to align the grid with external data or covariate layers, you can pass the `origin` argument as a tuple of two values (e.g., `c(0, 0)`). Similarly, you can specify the coordinate reference system (CRS) of your grid (`crs`).

```{r}
system.time(
grid_thin <- thin_points(
  data = sim_data,
  method = "grid",
  resolution = 1,
  origin = c(0, 0),
  crs = "epsg:4326",
  n = 1, # one point per grid cell
  trials = 10,
  all_trials = FALSE,
  seed = 123
))
nrow(largest(grid_thin))
```

Alternatively, you can pass a `terra::SpatRaster` object, and that grid will be used for the thinning process.

```{r, message=FALSE, warning=FALSE}
rast_obj <- terra::rast(xmin = -10, xmax = 10, ymin = -5, ymax = 5, res = 1)
system.time(
grid_raster_thin <- thin_points(
  data = sim_data,
  method = "grid", 
  raster_obj = rast_obj,
  trials = 10,
  n = 1,
  all_trials = FALSE,
  seed = 123
))
nrow(largest(grid_raster_thin))
```
```{r, echo = FALSE}
rast_obj <- terra::as.polygons(rast_obj)
rast_obj <- sf::st_as_sf(rast_obj)
ggplot() +
  geom_sf(data = rast_obj, color = "#353839") +
  geom_point(data = sim_data, aes(x = lon, y = lat),color = "#EB714B", alpha = 0.5) +
  geom_point(data = largest(grid_raster_thin), aes(x = lon, y = lat),color = "#5183B3", alpha = 1) +
  labs(title = "Grid-based thinning results",
       x = "Longitude",
       y = "Latitude",
       caption = "Blue points are kept, red points are deleted") +
  theme_minimal()
```

## Precision thinning

In this approach, coordinates are rounded to a certain precision to remove points that fall too close together. After removing points based on coordinate precision, the coordinate values are restored to their original locations. This is the simplest method and is more of a filtering when working with data from different sources with varying coordinate precision. To use it, you need to define the `precision` parameter, indicating the number of decimals to which the coordinates should be rounded.

```{r}
system.time(
precision_thin <- thin_points(
  data = sim_data,
  method = "precision", 
  precision = 0,
  trials = 50,
  all_trials = FALSE,
  seed = 123
))
nrow(largest(precision_thin))
```
```{r, echo = FALSE}
ggplot() +
  geom_point(data = sim_data, aes(x = lon, y = lat),  color = "#EB714B", size = 3, alpha = 0.5) +
  geom_point(data = largest(precision_thin), aes(x = lon, y = lat), color = "#5183B3", size = 3) +
  labs(title = "Precision-based thinning results",
       x = "Longitude",
       y = "Latitude",
       caption = "Blue points are kept, red points are deleted") +
  theme_minimal()
```

These are the methods implemented in **GeoThinneR**. Depending on your specific dataset and research needs, one method may be more suitable than others.

# Additional customization features

## Spatial thinning by group

In some cases, your dataset may include different groups, such as species, time periods, areas, or conditions, that you want to thin independently. The `group_col` parameter allows you to specify the column containing the grouping factor, and the thinning will be performed separately for each group. For example, in the simulated data where we have two species, we can use this parameter to thin each species independently:

```{r}
all_thin <- thin_points(
  data = sim_data,
  thin_dist = 100,
  seed = 123
)
grouped_thin <- thin_points(
  data = sim_data,
  group_col = "sp",
  thin_dist = 100,
  seed = 123
)

nrow(largest(all_thin))
nrow(largest(grouped_thin))
```
```{r, echo =FALSE, fig.show="hold",  out.width="50%"}
removed <- sim_data[!all_thin$retained[[1]], ]
removed_group <- sim_data[!grouped_thin$retained[[1]], ]

ggplot() +
  geom_point(data = removed, aes(x = lon, y = lat, shape = sp), color = "#EB714B",  alpha = 0.2) +
  geom_point(data = largest(all_thin), aes(x = lon, y = lat, shape = sp), color = "#5183B3", alpha = 1) +
  scale_shape_manual(values = c(sp1 = 15, sp2 = 17)) +
  ggtitle("Thinning without grouping") +
  theme_minimal()

ggplot() +
  geom_point(data = removed_group, aes(x = lon, y = lat, shape = sp), color = "#EB714B",  alpha = 0.2) +
  geom_point(data = largest(grouped_thin), aes(x = lon, y = lat, shape = sp), color = "#5183B3", alpha = 1) +
  scale_shape_manual(values = c(sp1 = 15, sp2 = 17)) +
  ggtitle("Thinning by group") +
  theme_minimal()
```



## Fixed number of points

What if you need to retain a fixed number of points that best covers the area where your data points are located, or for class balancing? The `target_points` parameter allows you to specify the number of points to keep, and the function will return that number of points spaced as separated as possible. Additionally, you can also set a `thin_dist` parameter so that no points closer than this distance will be retained. Currently, this approach is only implemented using the brute force method (`method = "distance"` and `search_type = "brute"`), so be cautious when applying it to very large datasets.

```{r}
targeted_thin <- thin_points(
  data = caretta,
  lon_col = "decimalLongitude",
  lat_col = "decimalLatitude",
  target_points = 150,
  thin_dist = 30,
  all_trials = FALSE,
  seed = 123,
  verbose = TRUE
)
nrow(largest(targeted_thin))
```
```{r, echo = FALSE}
ggplot() +
  geom_sf(data = medit, color = "#353839", fill = "antiquewhite", alpha = 0.7)  +
  geom_point(data = largest(targeted_thin), aes(x = decimalLongitude, y = decimalLatitude),color = "#5183B3", alpha = 1) +
  xlab("Longitude") + ylab("Latitude") +
  ggtitle("Exact number of occurrences") +
  theme(
      panel.grid.major = element_line(color = gray(.5),linetype = "dashed",linewidth = 0.5),
      panel.background = element_rect(fill = "white"),
      axis.title = element_blank(),
      legend.position = "bottom"
    )
```

## Select points by priority

In some scenarios, you may want to prioritize certain points based on a specific variable, such as coordinate uncertainty, data quality, or any other variable. The `priority` parameter allows you to assign a numeric weight to each point, where higher values indicate higher importance, and these points are more likely to be retained during thinning. This feature is available for the `"distance"`, `"grid"` and `"precision"` methods.

For example, in the sea turtle data downloaded from GBIF, there is a column named `coordinateUncertaintyInMeters`. To prioritize records with lower uncertainty, you can invert the values so that smaller uncertainties receive higher priority. A simple way to do this is:

```{r}
# Invert uncertainty to give more priority to lower uncertainty records
priority <- -caretta$coordinateUncertaintyInMeters

# For example, handle NAs by assigning lowest possible priority
priority[is.na(priority)] <- min(priority, na.rm = TRUE) - 1
```

Then apply thinning using the priority weights:

```{r}
# Standard grid thinning (no priority)
grid_thin <- thin_points(
  data = caretta,
  lon_col = "decimalLongitude",
  lat_col = "decimalLatitude",
  method = "grid",
  resolution = 2,
  seed = 123
)

# Grid thinning with priority
priority_thin <- thin_points(
  data = caretta,
  lon_col = "decimalLongitude",
  lat_col = "decimalLatitude",
  method = "grid",
  resolution = 2,
  priority = priority,
  seed = 123
)

mean(largest(grid_thin)$coordinateUncertaintyInMeters, na.rm = TRUE)
mean(largest(priority_thin)$coordinateUncertaintyInMeters, na.rm = TRUE)
```

**Notes**

* Higher `priority` values mean the point is more likely to be retained.
* In `"grid"` and `"precision"` methods, when multiple points fall into the same grid cell or rounded coordinate, the one with highest priority is kept.
* In the `"distance"` method, priority is used to resolve tie situations where multiple points have the same spatial influence; the point with lowest priority is dropped.
* If multiple points have the same priority, one is selected at random.
* By default, if provided `NA` values in `priority`, they are treated as the lowest possible priority.

# Other Packages

The **GeoThinneR** package was inspired by the work of many others who have developed methods and packages for working with spatial data and thinning techniques. Our goal with **GeoThinneR** is to offer additional flexibility in method selection and to address specific needs we encountered while using other packages. We would like to mention other tools that may be suitable for your work:

- **spThin**: The `thin()` function provides a brute-force spatial thinning of data, maximizing the number of retained points through random iterative repetitions, using the Haversine distance.
- **enmSdmX**: Includes the `geoThin()` function, which calculates all pairwise distances between points for thinning purposes.
- **dismo**: The `gridSample()` function samples points using a grid as stratification, providing an efficient method for spatial thinning.

# References

- Aiello‐Lammens, M. E., Boria, R. A., Radosavljevic, A., Vilela, B., and Anderson, R. P. (2015). *spThin: an R package for spatial thinning of species occurrence records for use in ecological niche models*. Ecography, 38(5), 541-545. https://doi.org/10.1111/ecog.01132

- Elseberg, J., Magnenat, S., Siegwart, R., & Nüchter, A. (2012). *Comparison of nearest-neighbor-search strategies and implementations for efficient shape registration*. Journal of Software Engineering for Robotics, 3(1), 2-12. https://hdl.handle.net/10446/86202

- Hijmans, R.J., Phillips, S., Leathwick, J., & Elith, J. (2023). *dismo: Species Distribution Modeling*. R Package Version 1.3-14. https://cran.r-project.org/package=dismo

- Mestre-Tomás, J. (2026). GeoThinneR: An R Package for Efficient Spatial Thinning of Species Occurrences and Point Data. *The R Journal, 18*(1), 299-314. https://doi.org/10.32614/RJ-2026-006

- Smith, A. B., Murphy, S. J., Henderson, D., & Erickson, K. D. (2023). *Including imprecisely georeferenced specimens improves accuracy of species distribution models and estimates of niche breadth*. Global Ecology and Biogeography, 32(3), 342-355. https://doi.org/10.1111/geb.13628

# Session info

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{estimate_k_max}
\alias{estimate_k_max}
\title{Estimate Maximum Neighbors for kd-Tree Thinning}
\usage{
estimate_k_max(coordinates, thin_dist, distance = c("haversine", "euclidean"))
}
\arguments{
\item{coordinates}{A matrix of spatial coordinates with two columns for longitude and latitude.}

\item{thin_dist}{A positive numeric value representing the thinning distance in kilometers.
This defines the resolution of the grid used for density calculations.}

\item{distance}{Distance metric used `c("haversine", "euclidean")`.}
}
\value{
A numeric value representing the maximum k (number of nearest neighbors) required
        for the densest regions in the dataset.
}
\description{
This function estimates the maximum value of k (the number of nearest neighbors)
for kd-tree-based thinning by evaluating the densest regions of a spatial dataset.
The function uses a histogram-based binning approach for efficiency and low memory usage.
}
\details{
The function divides the spatial domain into grid cells based on the specified thinning distance.
Grid cell sizes are determined assuming approximately 111.32 km per degree (latitude/longitude).
The function identifies the densest grid cells and their immediate neighbors to compute the maximum k value.
}
\examples{
# Generate sample data
set.seed(123)
coordinates <- matrix(runif(200, min = -10, max = 10), ncol = 2)

# Estimate k for kd-tree thinning
k_max <- estimate_k_max(coordinates, thin_dist = 50)
print(k_max)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance_thinning.R
\name{compute_neighbors_local_kdtree}
\alias{compute_neighbors_local_kdtree}
\title{Compute Neighbors Using Local kd-Trees}
\usage{
compute_neighbors_local_kdtree(
  coordinates,
  thin_dist,
  distance = c("haversine", "euclidean"),
  R = 6371,
  n_cores = 1
)
}
\arguments{
\item{coordinates}{A matrix of coordinates to thin, with two columns representing longitude and latitude.}

\item{thin_dist}{A positive numeric value representing the thinning distance in kilometers.}

\item{distance}{A character string specifying the distance metric to use `c("haversine", "euclidean")`.}

\item{R}{A numeric value representing the radius of the Earth in kilometers. The default is 6371 km.}

\item{n_cores}{An integer specifying the number of cores to use for parallel processing. The default is 1.}
}
\value{
A list where each element corresponds to a point and contains the indices of its neighbors, excluding the point itself.
}
\description{
Divides the search area into a grid of local regions and constructs kd-trees
for each region to compute neighbors efficiently. Neighbor regions are
also considered to ensure a complete search.
}
\examples{
set.seed(123)
coords <- matrix(runif(20, min = -180, max = 180), ncol = 2)

# Compute neighbors using local kd-trees with Euclidean distance
neighbors <- compute_neighbors_local_kdtree(coords, thin_dist = 10, n_cores = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{calculate_spatial_coverage}
\alias{calculate_spatial_coverage}
\title{Calculate Spatial Coverage (Convex Hull Area)}
\usage{
calculate_spatial_coverage(coordinates, distance = "haversine")
}
\arguments{
\item{coordinates}{A matrix of coordinates (longitude and latitude or planar x/y).}

\item{distance}{A character string: "haversine" (default) or "euclidean".}
}
\value{
A numeric value representing the convex hull area (km2 or unit2).
}
\description{
Computes the area of the convex hull formed by the points.
Uses geodetic area (km2) if coordinates are lon/lat and distance = "haversine",
otherwise computes area in squared map units.
}
\examples{
# Geographic coordinates (lon/lat)
set.seed(456)
coords_geo <- matrix(cbind(runif(10, -10, 10), runif(10, 40, 50)), ncol = 2)
area_haversine <- calculate_spatial_coverage(coords_geo, distance = "haversine")
print(round(area_haversine, 2))  # in km2

# Projected coordinates (Euclidean/map units)
coords_proj <- matrix(runif(20), ncol = 2) * 100  # e.g., map units
area_euclidean <- calculate_spatial_coverage(coords_proj, distance = "euclidean")
print(round(area_euclidean, 2))  # in unit2

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{thunnus}
\alias{thunnus}
\title{Yellowfin Tuna (\emph{Thunnus albacares}) Worldwide Occurrence Records}
\format{
A data frame with 80,163 rows and 3 variables:
\describe{
  \item{decimalLongitude}{Numeric. Longitude coordinates (WGS84).}
  \item{decimalLatitude}{Numeric. Latitude coordinates (WGS84).}
  \item{year}{Integer. Year in which the observation was recorded.}
}
}
\source{
Global Biodiversity Information Facility (GBIF) \cr
GBIF Occurrence Download https://doi.org/10.15468/dl.xsyrkh \cr
Accessed from R via \pkg{rgbif} (https://github.com/ropensci/rgbif) on 2025-02-07
}
\usage{
data("thunnus")
}
\description{
This dataset contains georeferenced occurrence records of the Yellowfin Tuna
(\emph{Thunnus albacares}) from 1950 to 2025, obtained globally from the Global Biodiversity
Information Facility (GBIF). The dataset was filtered to include presence-only records
with valid geographic coordinates and year, and duplicate coordinates were removed.
}
\details{
The data were accessed through the GBIF API and used for benchmarking thinning algorithms
in the GeoThinneR package. Coordinates include marine records, and
the data were globally sampled across many decades.

For reproducibility, the GBIF download script is included in \code{data-raw/thunnus_download.R}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{lon_lat_to_cartesian}
\alias{lon_lat_to_cartesian}
\title{Convert Geographic Coordinates to Cartesian Coordinates}
\usage{
lon_lat_to_cartesian(lon, lat, R = 6371)
}
\arguments{
\item{lon}{Numeric vector of longitudes in degrees.}

\item{lat}{Numeric vector of latitudes in degrees.}

\item{R}{Radius of the Earth in kilometers (default: 6371 km).}
}
\value{
A numeric matrix with three columns (x, y, z) representing Cartesian coordinates.
}
\description{
This function converts geographic coordinates, given as longitude and latitude in degrees, to Cartesian coordinates (x, y, z) assuming a spherical Earth model.
}
\examples{
lon <- c(-122.4194, 0)
lat <- c(37.7749, 0)
lon_lat_to_cartesian(lon, lat)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thinning_algorithm.R
\name{select_target_points}
\alias{select_target_points}
\title{Select Target Number of Points for Spatial Thinning}
\usage{
select_target_points(
  distance_matrix,
  target_points,
  thin_dist,
  trials,
  all_trials = FALSE
)
}
\arguments{
\item{distance_matrix}{A matrix of pairwise distances between points.}

\item{target_points}{An integer specifying the number of points to retain.}

\item{thin_dist}{A positive numeric value representing the thinning distance in kilometers.}

\item{trials}{A positive integer specifying the number of thinning trials to perform. Default is 10.}

\item{all_trials}{A logical value indicating whether to return results of all attempts (`TRUE`) or only the best attempt with the most points retained (`FALSE`). Default is `FALSE`.}
}
\value{
A list of logical vectors indicating which points are kept in each trial if `all_trials` is `TRUE`; otherwise, a list with a single logical vector indicating the points kept in the best trial.
}
\description{
This function selects a specified number of points from a spatial dataset while maximizing
the distance between selected points.
}
\examples{
# Example distance matrix (3 points)
dist_matrix <- matrix(c(0, 2, 5,
                        2, 0, 3,
                        5, 3, 0), ncol = 3)

# Select 2 points maximizing distance
result <- select_target_points(dist_matrix, target_points = 2,
                              thin_dist = 4, trials = 5, all_trials = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thin_points.R
\name{thin_points}
\alias{thin_points}
\title{Spatial Thinning of Points}
\usage{
thin_points(
  data,
  lon_col = "lon",
  lat_col = "lat",
  group_col = NULL,
  method = c("distance", "grid", "precision"),
  trials = 10,
  all_trials = FALSE,
  seed = NULL,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{data}{A data frame or tibble containing the input points to thin. Must contain longitude and latitude columns.}

\item{lon_col}{Character name of the column with longitude coordinates (default: `"lon"`).}

\item{lat_col}{Character name of the column with latitude coordinates (default: `"lat"`).}

\item{group_col}{Character name of the column for grouping points (e.g., species name, year). If `NULL`, no grouping is applied.}

\item{method}{Thinning method to use. One of `"distance", "grid", "precision"`.}

\item{trials}{Number of thinning iterations to perform (default: `10`). Must be a positive integer.}

\item{all_trials}{If `TRUE`, returns results of all attempts; if `FALSE`, returns the best attempt with the most points retained (default: `FALSE`).}

\item{seed}{Optional; an integer seed for reproducibility of results.}

\item{verbose}{If `TRUE`, prints progress messages (default: `FALSE`).}

\item{...}{Additional arguments passed to specific thinning methods. See Details.}
}
\value{
A `GeoThinned` object (S3 class), which contains:
\itemize{
  \item `retained`: A list of logical vectors (one per trial) indicating retained points.
  \item `original_data`: The original input dataset.
  \item `method`: The thinning method used.
  \item `params`: A list of the thinning parameters used.
}
By default, `thin_points()` returns only the trial with the largest number of retained points.
To access all thinning trials, set `all_trials = TRUE`; otherwise, functions such as `largest()` and `get_trial()` will always refer to the same subset.
}
\description{
This function performs spatial thinning of geographic points to reduce
point density while maintaining spatial representation. Points are thinned
based on a specified distance, grid, or decimal precision, with support for
multiple trials and optional grouping.
By default, only the largest subset is returned. To return all trials, set `all_trials = TRUE`.
}
\details{
The following thinning methods are available:

\describe{
 \item{`"distance"`}{Forces a specific minimum distance between points.}
 \item{`"grid"`}{Applies a grid-based thinning method.}
 \item{`"precision"`}{Utilizes precision-based thinning.}
}

\strong{Distance-based thinning}

The specific parameters for distance-based thinning are:

\describe{
 \item{`thin_dist`}{A positive numeric value representing the thinning distance in kilometers.}
 \item{`search_type`}{A character string indicating the neighbor search method 'c("local_kd_tree", "k_estimation", "kd_tree", "brute")'. The defult value is 'local_kd_tree'.}
 \item{`distance`}{Distance metric to use 'c("haversine", "euclidean")'. Default is Haversine for geographic coordinates.}
 \item{`R`}{The radius of the Earth in kilometers. Default is 6371 km.}
 \item{`target_points`}{Optional integer specifying the number of points to retain. If 'NULL' (default), the function tries to maximize the number of points retained.}
 \item{`n_cores`}{Number of cores for parallel processing (only for '"local_kd_tree"'). Default is 1.}
}

\strong{Grid-based thinning}

The specific parameters for grid-based thinning are:

\describe{
 \item{`thin_dist`}{A positive numeric value representing the thinning distance in kilometers.}
 \item{`resolution`}{A numeric value representing the resolution (in degrees) of the raster grid. If provided, this takes priority over 'thin_dist'.}
 \item{`origin`}{A numeric vector of length 2 (e.g., 'c(0, 0)'), specifying the origin of the raster grid (optional).}
 \item{`raster_obj`}{An optional 'terra::SpatRaster' object to use for grid thinning. If provided, the raster object will be used instead of creating a new one.}
 \item{`n`}{A positive integer specifying the maximum number of points to retain per grid cell (default: 1).}
 \item{`crs`}{An optional CRS (Coordinate Reference System) to project the coordinates and raster (default WGS84, 'epsg:4326'). This can be an EPSG code, a PROJ.4 string, or a 'terra::crs' object.}
 \item{`priority`}{A numeric vector of the same length as the number of points with numerical values indicating the priority of each point. Instead of eliminating points randomly, higher values are preferred during thinning.}
}

\strong{Precision-based thinning}

The specific parameters for precision-based thinning are:

\describe{
 \item{`precision`}{A positive integer specifying the number of decimal places to which coordinates should be rounded. Default is 4.}
 \item{`priority`}{A numeric vector of the same length as the number of points with numerical values indicating the priority of each point. Instead of eliminating points randomly, higher values are preferred during thinning.}
}

For more information on specific thinning methods and inputs, refer to their respective documentation:

\itemize{
 \item `distance_thinning()`
 \item `grid_thinning()`
 \item `precision_thinning()`
}
}
\examples{
# Basic usage
set.seed(123)
sample_data <- data.frame(
  lon = runif(100, -10, 10),
  lat = runif(100, -5, 5)
)

result <- thin_points(sample_data, method = "distance", thin_dist = 100)

# Grouped thinning
sample_data$species <- sample(c("A", "B"), 100, replace = TRUE)
grouped_result <- thin_points(sample_data, group_col = "species",
                              method = "distance", thin_dist = 100)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid_thinning.R
\name{grid_thinning}
\alias{grid_thinning}
\title{Perform Grid-Based Thinning of Spatial Points}
\usage{
grid_thinning(
  coordinates,
  thin_dist = NULL,
  resolution = NULL,
  origin = NULL,
  raster_obj = NULL,
  n = 1,
  trials = 10,
  all_trials = FALSE,
  crs = "epsg:4326",
  priority = NULL
)
}
\arguments{
\item{coordinates}{A numeric matrix or data frame with two columns representing the x (longitude) and y (latitude) coordinates of the points.}

\item{thin_dist}{A numeric value representing the thinning distance in kilometers. It will be converted to degrees if `resolution` is not provided.}

\item{resolution}{A numeric value representing the resolution (in degrees) of the raster grid. If provided, this takes priority over `thin_dist`.}

\item{origin}{A numeric vector of length 2 (e.g., `c(0, 0)`), specifying the origin of the raster grid (optional).}

\item{raster_obj}{An optional `terra::SpatRaster` object to use for grid thinning. If provided, the raster object will be used instead of creating a new one.}

\item{n}{A positive integer specifying the maximum number of points to retain per grid cell (default: 1).}

\item{trials}{An integer specifying the number of trials to perform for thinning (default: 10).}

\item{all_trials}{A logical value indicating whether to return results for all trials (`TRUE`) or just the first trial (`FALSE`, default).}

\item{crs}{An optional CRS (Coordinate Reference System) to project the coordinates and raster (default WGS84, `epsg:4326`). This can be an EPSG code, a PROJ.4 string, or a `terra::crs` object.}

\item{priority}{A numeric vector of the same length as the number of points, specifying a priority weight for each point. Higher values indicate higher importance and are favored when selecting which points to retain. Priority is used to guide selection when multiple candidate points are otherwise equally valid (e.g., points in the same grid cell, with the same rounded coordinates, or with the same number of neighbors).}
}
\value{
A list of logical vectors indicating which points to keep for each trial.
}
\description{
This function performs thinning of spatial points by assigning them to grid cells based on a specified resolution or thinning distance. It can either create a new raster grid or use an existing `terra::SpatRaster` object.
}
\examples{
# Example: Grid thinning using thin_dist
coords <- matrix(c(-122.4194, 37.7749,
                        -122.4195, 37.7740,
                        -122.4196, 37.7741), ncol = 2, byrow = TRUE)

result <- grid_thinning(coords, thin_dist = 10, trials = 5, all_trials = TRUE)
print(result)

# Example: Grid thinning using a custom resolution
result_res <- grid_thinning(coords, resolution = 0.01, n = 2, trials = 5)
print(result_res)

# Example: Using a custom raster object
library(terra)
rast_obj <- terra::rast(nrows = 100, ncols = 100, xmin = -123, xmax = -121, ymin = 36, ymax = 38)
result_raster <- grid_thinning(coords, raster_obj = rast_obj, trials = 5)
print(result_raster)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{caretta}
\alias{caretta}
\title{Loggerhead Sea Turtle (\emph{Caretta caretta}) Occurrence Records in the Mediterranean Sea}
\format{
A data frame with 8340 rows and 5 variables:
\describe{
  \item{decimalLongitude}{Numeric. Longitude coordinates (WGS84).}
  \item{decimalLatitude}{Numeric. Latitude coordinates (WGS84).}
  \item{year}{Integer. Year in which the observation was recorded.}
  \item{species}{Character. The species name (\emph{Caretta caretta}).}
  \item{coordinateUncertaintyInMeters}{Numeric. Positional uncertainty of the coordinates, in meters.}
}
}
\source{
Global Biodiversity Information Facility (GBIF) \cr
GBIF Occurrence Download https://doi.org/10.15468/dl.9jcjrm \cr
Accessed from R via \pkg{rgbif} (https://github.com/ropensci/rgbif) on 2025-02-07
}
\usage{
data("caretta")
}
\description{
This dataset contains georeferenced occurrence records of the Loggerhead Sea Turtle
(\emph{Caretta caretta}) within the Mediterranean Sea from 1962 to 2025.
The data were obtained from the Global Biodiversity Information Facility (GBIF)
and filtered to include only marine points falling within a defined Mediterranean polygon.
Duplicated coordinates were removed.
}
\details{
The data were accessed through the GBIF API and clipped spatially using
a Mediterranean Sea polygon (https://www.marineregions.org/gazetteer.php?p=details&id=1905).
All records were filtered to ensure presence-only data with valid geographic coordinates.
Duplicate coordinate pairs were removed to reduce redundancy.

This dataset is used as a real-world example throughout the GeoThinneR package and the accompanying article.

For reproducibility, the GBIF download script is included in \code{data-raw/caretta_download.R}.

The Mediterranean Sea polygon used to spatially filter the records was download from Marineregions.org:
\url{https://www.marineregions.org/gazetteer.php?p=details&id=1905}. It is distributed with the package as
\code{inst/extdata/mediterranean_sea.gpkg} and used in the \code{data-raw/caretta_download.R} script.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{is_lonlat}
\alias{is_lonlat}
\title{Check for Longitude/Latitude Coordinates}
\usage{
is_lonlat(lon, lat, tolerance = 0.1)
}
\arguments{
\item{lon}{Numeric vector of longitudes in degrees.}

\item{lat}{Numeric vector of latitudes in degrees.}

\item{tolerance}{Numeric tolerance (in degrees) for checking the global
longitude/latitude bounds. Default is 0.1.}
}
\value{
A logical value. `TRUE` if values are within the ranges and `FALSE` otherwise.
}
\description{
This function checks whether a pair of coordinate vectors represent
geographic longitude and latitude values. The function returns `TRUE`
if all longitude values are within -180 (- tolerance) and 180 (+ tolerance)
and the latitude is within -90 (- tolerance) and 90 (+ tolerance).
}
\examples{
is_lonlat(lon = c(-3, 10, 179), lat = c(40, -20, 5))
is_lonlat(lon = c(100000, 150000), lat = c(4500000, 4600000))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thinning_algorithm.R
\name{max_thinning_algorithm}
\alias{max_thinning_algorithm}
\title{Thinning Algorithm for Spatial Data}
\usage{
max_thinning_algorithm(
  neighbor_indices,
  trials,
  all_trials = FALSE,
  priority = NULL
)
}
\arguments{
\item{neighbor_indices}{A list of integer vectors where each element contains the indices of the neighboring points for each point in the dataset.}

\item{trials}{A positive integer specifying the number of thinning trials to perform. Default is 10.}

\item{all_trials}{A logical value indicating whether to return results of all attempts (`TRUE`) or only the best attempt with the most points retained (`FALSE`). Default is `FALSE`.}

\item{priority}{A numeric vector of the same length as the number of points, specifying a priority weight for each point. Higher values indicate higher importance and are favored when selecting which points to retain. Priority is used to guide selection when multiple candidate points are otherwise equally valid (e.g., points in the same grid cell, with the same rounded coordinates, or with the same number of neighbors).}
}
\value{
A list of logical vectors indicating which points are kept in each trial if all_trials is TRUE; otherwise, a list with a single logical vector indicating the points kept in the best trial.
}
\description{
This function performs the core thinning algorithm used to reduce the density of points
in spatial data while maintaining spatial representation. It iteratively removes the
points with the most neighbors until no points with neighbors remain. The algorithm
supports multiple trials to find the optimal thinning solution.
}
\examples{
# Example usage within a larger thinning function
neighbor_indices <- list(c(2, 3), c(1, 3), c(1, 2))
trials <- 5
all_trials <- FALSE
kept_points <- max_thinning_algorithm(neighbor_indices, trials, all_trials)
print(kept_points)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GeoThinned_class.R
\name{as_GeoThinned}
\alias{as_GeoThinned}
\alias{new_GeoThinned}
\alias{print.GeoThinned}
\alias{summary.GeoThinned}
\alias{print.summary.GeoThinned}
\alias{plot.GeoThinned}
\alias{largest}
\alias{largest.GeoThinned}
\alias{largest_index}
\alias{largest_index.GeoThinned}
\alias{get_trial}
\alias{get_trial.GeoThinned}
\alias{as_sf}
\alias{as_sf.GeoThinned}
\title{GeoThinned Object Constructor and Methods}
\usage{
as_GeoThinned(retained, method, params = list(), original_data = NULL)

new_GeoThinned(retained, method, params = list(), original_data = NULL)

\method{print}{GeoThinned}(x, ...)

\method{summary}{GeoThinned}(object, trial = NULL, ...)

\method{print}{summary.GeoThinned}(x, ...)

\method{plot}{GeoThinned}(
  x,
  trial = NULL,
  show_original = TRUE,
  col_original = "#EB714B",
  col_thinned = "#5183B3",
  pch_original = 1,
  pch_thinned = 16,
  main = NULL,
  ...
)

largest(x, ...)

\method{largest}{GeoThinned}(x, ...)

largest_index(x, ...)

\method{largest_index}{GeoThinned}(x, ...)

get_trial(x, trial = NULL, ...)

\method{get_trial}{GeoThinned}(x, trial = NULL, ...)

as_sf(x, ...)

\method{as_sf}{GeoThinned}(x, trial = NULL, crs = 4326, ...)
}
\arguments{
\item{retained}{A list of logical vectors indicating retained points per trial.}

\item{method}{The thinning method used (e.g., "distance", "grid", "precision").}

\item{params}{A list of parameters used in thinning.}

\item{original_data}{The original unmodified data.}

\item{x}{An object of class \code{GeoThinned}.}

\item{...}{Additional arguments (ignored).}

\item{object}{An object of class \code{GeoThinned}.}

\item{trial}{Integer index of the thinning trial to extract (for \code{summary()}, \code{get_trial()}, \code{plot()}, \code{as_sf()}). Default `NULL`, which will return the largest dataset.}

\item{show_original}{Logical, whether to show original points.}

\item{col_original}{Colors for original points.}

\item{col_thinned}{Colors for thinned points.}

\item{pch_original}{Point shapes for original points.}

\item{pch_thinned}{Point shapes for thinned points.}

\item{main}{Title of the plot.}

\item{crs}{Coordinate reference system to assign to the resulting \code{sf} object (optional).}
}
\value{
A \code{GeoThinned} object or associated results (summary, plot, trial subset).
When `thin_points()` is run with `all_trials = FALSE`, the returned object contains only the largest trial; therefore all methods refer to this single subset.
}
\description{
Create and interact with spatial thinning results stored in a GeoThinned object.
}
\details{
When the 's2' package is not installed or the installed version is lower than 1.1.0,
the spatial coverage metric for geographic coordinates from the \code{summary()} is not computed.
This is an optional summary metric and does not affect the thinning process or any other package functionality.
}
\seealso{
\code{\link{thin_points}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{compute_nearest_neighbor_distances}
\alias{compute_nearest_neighbor_distances}
\title{Compute Nearest Neighbor Distances}
\usage{
compute_nearest_neighbor_distances(
  coordinates,
  distance = "haversine",
  R = 6371
)
}
\arguments{
\item{coordinates}{A matrix of coordinates with two columns.}

\item{distance}{A character string: "haversine" (default) or "euclidean".}

\item{R}{Radius of the Earth in kilometers. Default is 6371.}
}
\value{
A numeric vector of nearest neighbor distances, in meters (haversine) or in map units (euclidean).
}
\description{
Calculates nearest neighbor distances using geodesic or Euclidean distance.
}
\examples{
# Example with geographic (longitude/latitude) coordinates
set.seed(123)
coords_geo <- matrix(cbind(runif(10, -10, 10), runif(10, 40, 50)), ncol = 2)
nnd_haversine <- compute_nearest_neighbor_distances(coords_geo, distance = "haversine")
print(round(nnd_haversine, 2))  # in km

# Example with projected coordinates (Euclidean)
coords_proj <- matrix(runif(20), ncol = 2) * 100  # e.g., meters or map units
nnd_euclidean <- compute_nearest_neighbor_distances(coords_proj, distance = "euclidean")
print(round(nnd_euclidean, 2))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance_thinning.R
\name{compute_neighbors_kdtree}
\alias{compute_neighbors_kdtree}
\title{Compute Neighbors Using kd-Tree}
\usage{
compute_neighbors_kdtree(
  coordinates,
  thin_dist,
  k = NULL,
  distance = c("haversine", "euclidean"),
  R = 6371
)
}
\arguments{
\item{coordinates}{A matrix of coordinates to thin, with two columns representing longitude and latitude.}

\item{thin_dist}{A positive numeric value representing the thinning distance in kilometers.}

\item{k}{An integer specifying the maximum number of neighbors to consider for each point.}

\item{distance}{A character string specifying the distance metric to use `c("haversine", "euclidean")`.}

\item{R}{A numeric value representing the radius of the Earth in kilometers. The default is 6371 km.}
}
\value{
A list where each element corresponds to a point and contains the indices of its neighbors, excluding the point itself.
}
\description{
Computes neighbors for each point in a set of coordinates using a kd-tree
for efficient neighbor searches. This method is particularly useful for large datasets.
}
\details{
This function uses kd-tree (via `nabor` package) for efficient spatial searches. The kd-tree inherently works with Euclidean distances.
If `"haversine"` is selected, the function first converts geographic coordinates to 3D Cartesian coordinates before constructing the kd-tree.
}
\examples{
set.seed(123)
coords <- matrix(runif(20, min = -180, max = 180), ncol = 2)

# Compute neighbors using kd-tree
neighbors <- compute_neighbors_kdtree(coords, thin_dist = 10,)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance_thinning.R
\name{distance_thinning}
\alias{distance_thinning}
\title{Perform Distance-Based Thinning}
\usage{
distance_thinning(
  coordinates,
  thin_dist = 10,
  trials = 10,
  all_trials = FALSE,
  search_type = c("local_kd_tree", "k_estimation", "kd_tree", "brute"),
  target_points = NULL,
  priority = NULL,
  distance = c("haversine", "euclidean"),
  R = 6371,
  n_cores = 1
)
}
\arguments{
\item{coordinates}{A matrix of coordinates to thin, with two columns representing longitude and latitude.}

\item{thin_dist}{A positive numeric value representing the thinning distance in kilometers.}

\item{trials}{An integer specifying the number of trials to run for thinning. Default is 10.}

\item{all_trials}{A logical indicating whether to return results of all attempts (`TRUE`) or only the best attempt with the most points retained (`FALSE`). Default is `FALSE`.}

\item{search_type}{A character string indicating the neighbor search method `c("local_kd_tree", "k_estimation", "kd_tree", "brute")`. The default value is `local_kd_tree`. See details.}

\item{target_points}{Optional integer specifying the number of points to retain. If `NULL` (default), the function tries to maximize the number of points retained.}

\item{priority}{A numeric vector of the same length as the number of points, specifying a priority weight for each point. Higher values indicate higher importance and are favored when selecting which points to retain. Priority is used to guide selection when multiple candidate points are otherwise equally valid (e.g., points in the same grid cell, with the same rounded coordinates, or with the same number of neighbors).}

\item{distance}{Distance metric to use `c("haversine", "euclidean")`. Default is Haversine for geographic coordinates.}

\item{R}{Radius of the Earth in kilometers (default: 6371 km).}

\item{n_cores}{Number of cores for parallel processing (only for `"local_kd_tree"`). Default is 1.}
}
\value{
A list. If `all_trials` is `FALSE`, the list contains a single logical vector indicating which points are kept in the best trial. If `all_trials` is `TRUE`, the list contains a logical vector for each trial.
}
\description{
This function applies a distance-based thinning algorithm using a kd-tree or brute-force approach.
Two modified algorithms based on kd-trees (local kd-trees and estimating the maximum number of neighbors) are implemented which scale better for large datasets.
The function removes points that are closer than a specified distance to each other while maximizing spatial representation.
}
\details{
- `"kd_tree"`: Uses a single kd-tree for efficient nearest-neighbor searches.
- `"local_kd_tree"`: Builds multiple smaller kd-trees for better scalability.
- `"k_estimation"`: Approximates a maximum number of neighbors per point to reduce search complexity.
- `"brute"`: Computes all pairwise distances (inefficient for large datasets).
}
\examples{
# Generate sample coordinates
set.seed(123)
result  <- matrix(runif(20, min = -180, max = 180), ncol = 2) # 10 random points

# Perform thinning with local kd-trees
result_partitioned <- distance_thinning(result , thin_dist = 5000, trials = 5,
                                       search_type = "local_kd_tree", all_trials = TRUE)
print(result_partitioned)

# Perform thinning estimating max number of neighbors
result_estimated <- distance_thinning(result , thin_dist = 5000, trials = 5,
                                       search_type = "k_estimation", all_trials = TRUE)
print(result_estimated)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/precision_thinning.R
\name{precision_thinning}
\alias{precision_thinning}
\title{Precision Thinning of Spatial Points}
\usage{
precision_thinning(
  coordinates,
  precision = 4,
  trials = 10,
  all_trials = FALSE,
  priority = NULL
)
}
\arguments{
\item{coordinates}{A numeric matrix or data frame with two columns representing the longitude and latitude of points.}

\item{precision}{A positive integer specifying the number of decimal places to which coordinates should be rounded. Default is 4.}

\item{trials}{A positive integer specifying the number of thinning trials to perform. Default is 10.}

\item{all_trials}{A logical value indicating whether to return results for all trials (`TRUE`) or just the first/best trial (`FALSE`). Default is `FALSE`.}

\item{priority}{A numeric vector of the same length as the number of points, specifying a priority weight for each point. Higher values indicate higher importance and are favored when selecting which points to retain. Priority is used to guide selection when multiple candidate points are otherwise equally valid (e.g., points in the same grid cell, with the same rounded coordinates, or with the same number of neighbors).}
}
\value{
If `all_trials` is `FALSE`, returns a logical vector indicating which points were kept in the first trial.
If `all_trials` is `TRUE`, returns a list of logical vectors, one for each trial.
}
\description{
This function performs thinning of spatial points by rounding their coordinates to a specified precision and removing duplicates.
It can perform multiple trials of this process and return the results for all or just the best trial.
}
\details{
The function performs multiple trials to account for randomness in the order of point selection. By default, it returns the first trial,
but setting `all_trials = TRUE` will return the results of all trials.
}
\examples{
# Example usage
coords <- matrix(c(-123.3656, 48.4284, -123.3657, 48.4285, -123.3658, 48.4286), ncol = 2)
result <- precision_thinning(coords, precision = 3, trials = 5, all_trials = TRUE)
print(result)

# Example with a single trial and lower precision
result_single <- precision_thinning(coords, precision = 2, trials = 1, all_trials = FALSE)
print(result_single)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance_thinning.R
\name{compute_neighbors_brute}
\alias{compute_neighbors_brute}
\title{Compute Neighbors Using Brute-Force}
\usage{
compute_neighbors_brute(
  coordinates,
  thin_dist,
  distance = c("haversine", "euclidean"),
  R = 6371
)
}
\arguments{
\item{coordinates}{A matrix of coordinates to thin, with two columns representing longitude and latitude.}

\item{thin_dist}{A positive numeric value representing the thinning distance in kilometers.}

\item{distance}{A character string specifying the distance metric to use `c("haversine", "euclidean")`.}

\item{R}{A numeric value representing the radius of the Earth in kilometers. The default is 6371 km.}
}
\value{
A list where each element corresponds to a point and contains the indices of its neighbors.
}
\description{
Computes neighbors for each point in a set of coordinates using a greedy
approach. All pairwise distances are calculated to identify neighbors within
a specified distance threshold.
}
\examples{
set.seed(123)
coords <- matrix(runif(20, min = -180, max = 180), ncol = 2)

# Compute neighbors using brute fore
neighbors <- compute_neighbors_brute(coords, thin_dist = 10,)

}
